/*
=============================================================================
Smartport Virtual Hard Drive, version 1.00.0

Copyright 2010-2012 Cédric PELTIER, pcedric@hotmail.com

This file is part of the Smartport VHD project

Smartport VHD is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 3

Smartport VHD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Smartport VHD. If not, see <http://www.gnu.org/licenses/>.
=============================================================================
*/

#include "i2c.h"

/// Wait for the end of the current transmission
uint8_t i2c_sync()
{
	uint8_t l_rc;
   #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
   if(gb_debug & CFG_DEBUG_MASK_I2C)
   	puts_P(PSTR("I2C: i2c_sync()\r\n"));
   #endif
   uint16_t i = I2C_MAX_SYNC;
	while (!(TWCR & (1 << TWINT)) && i--!=0) _delay_us(10);	
	l_rc = (TWSR & 0xF8);
   #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
   if(gb_debug & CFG_DEBUG_MASK_I2C)
   	printf_P(PSTR("I2C: rc = %02X\r\n"),l_rc);
   #endif
	return l_rc;
}

/// Start a transmission
uint8_t i2c_start()
{
   #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
   if(gb_debug & CFG_DEBUG_MASK_I2C)
   	puts_P(PSTR("I2C: i2c_start()\r\n"));
   #endif
	TWCR = (1 << TWINT) | (1 << TWSTA) | (1 << TWEN);
	return i2c_sync();
}

/// Send data without ack
uint8_t i2c_nack()
{
   #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
   if(gb_debug & CFG_DEBUG_MASK_I2C)
   	puts_P(PSTR("I2C: i2c_nack()\r\n"));
   #endif
	TWCR = (1 << TWINT) | (1 << TWEN);
	return i2c_sync();
}

/// Send data with ack
uint8_t i2c_ack()
{
   #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
   if(gb_debug & CFG_DEBUG_MASK_I2C)
   	puts_P(PSTR("I2C: i2c_ack()\r\n"));
   #endif
	TWCR = (1 << TWEA) | (1 << TWINT) | (1 << TWEN);
	return i2c_sync();
}

/// Stop the current transmission
void i2c_stop()
{
   #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
   if(gb_debug & CFG_DEBUG_MASK_I2C)
   	puts_P(PSTR("I2C: i2c_stop()\r\n"));
   #endif
	TWCR = (1 << TWINT) | (1 << TWEN) | (1 << TWSTO);
}

/// Initialize the communication with the slave
bool i2c_init(uint8_t l_id, uint8_t l_addr, uint8_t l_rw)
{
  uint8_t i = 0;
  uint8_t l_rc,l_rc2;

   #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
   if(gb_debug & CFG_DEBUG_MASK_I2C)
   	printf_P(PSTR("I2C: i2c_init(%02X,%02X,%d)\r\n"),l_id,l_addr,l_rw);
   #endif

  // prescaler set to 1 
  TWSR = 0x00;
  // SCL frequency = 16000000 / (16 + 2 * 74 - 4) = 100KHz
  // SCL frequency = 16000000 / (16 + 2 * 154 - 4) = 50KHz
  TWBR = 100;
  // enable our pullups on SCL and SDA
  PORTD |= 3; 
  
  l_rc2 = false;
  while(i < I2C_MAX_TRIES && !l_rc2)
  {
  	  l_rc = TW_MT_ARB_LOST;
	  while (i++ < I2C_MAX_TRIES && l_rc == TW_MT_ARB_LOST)
	  	  l_rc = i2c_start();
	  if((l_rc != TW_START) && (l_rc != TW_REP_START))
	  	  return false;

	  TWDR = (l_id & 0xF0) | (l_addr & 0x07) | l_rw;
	  l_rc = i2c_nack();

	  // Check the TWSR status
	  l_rc2 = (l_rc != TW_MT_SLA_NACK) && (l_rc != TW_MT_ARB_LOST);
	  if (l_rc2 && l_rc != (l_rw==0?TW_MT_SLA_ACK:TW_MR_SLA_ACK))
	  {
			#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
			if(gb_debug & CFG_DEBUG_MASK_I2C)
				puts_P(PSTR("I2C: rc != TW_MT_SLA_ACK"));
			#endif
	  	  return false;
	  }
  }  
  return (i<I2C_MAX_TRIES);
}
  
/// Send a byte to the slave
bool i2c_write(uint8_t l_data)
{
  uint8_t l_rc;
  
   #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
   if(gb_debug & CFG_DEBUG_MASK_I2C)
   	printf_P(PSTR("I2C: i2c_write(%02X)\r\n"),l_data);
   #endif

  // Send the Data to I2C Bus
  TWDR = l_data;

  // Transmit I2C Data
  l_rc = i2c_nack();

  // Check the TWSR status
  return (l_rc == TW_MT_DATA_ACK);
}

/// Receive a byte from the slave
bool i2c_read(uint8_t *lp_data,bool l_ack)
{
  uint8_t l_rc;

   #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
   if(gb_debug & CFG_DEBUG_MASK_I2C)
   	printf_P(PSTR("I2C: i2c_read(%04X,%d)\r\n"),lp_data,l_ack);
   #endif
   
  if (l_ack)
  {
    // Read I2C Data and Send Acknowledge
    l_rc = i2c_ack();
    if (l_rc != TW_MR_DATA_ACK)
    	 return false;
  }
  else
  {
    // Read I2C Data and Send No Acknowledge
    l_rc = i2c_nack();
    if (l_rc != TW_MR_DATA_NACK)
    	 return false;
  }

  // Get the Data
  *lp_data = TWDR;
  return true;
}

