/**
=============================================================================
Smartport Virtual Hard Drive, version 0.99.0 (beta)
smartport (aka Cbus) protocol interpreter

If you would like to be involved in this project, or simply get my 
documentation for your own smartport project, feel free to contact me at
pcedric@hotmail.com

Copyright 2010 Cédric PELTIER, pcedric@hotmail.com

This file is part of the Smartport VHD project

Smartport VHD is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 3

Smartport VHD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Smartport VHD. If not, see <http://www.gnu.org/licenses/>.
=============================================================================
*/

/**
Timers configuration for sending and receiving

        Send setup              Receive setup
TIFR1   (1<<OCIE1C)|(1<<ICF1)   (1<<OCIE1C)|(1<<ICF1)
ICR1    63                      0
OCR1C   15                      0
TCNT1   0                       0
TCCR0A  0                       0
TCCR1A  (1<<COM1C0)|(1<<COM1C1)|(1<<WGM11)  0
TCCR1B  (1<<WGM13)|(1<<WGM12)   0
TIMSK1  0                       0

        Send start              Receive start
TCCR1B  |(1<<CS10)              (1<<CS11)
TIMSK1  (1<<OCIE1C)             (1<<ICIE1)

        Send stop               Receive stop
TCCR1B  0
TIMSK1  0                       

*/

#include "spvhd.h"
#include "protocol.h"
#include "cache.h"
#include "usb_storage.h"
#include "rwts.h"
#include <string.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#ifdef CFG_DEBUG_USB
#include "usb_serial.h"
#endif
#ifdef CFG_DEBUG_SERIAL
#if(LUFA_VERSION==100219)
#include <LUFA/Drivers/Peripheral/SerialStream.h>
#else
#include <LUFA/Drivers/Peripheral/Serial.h>
#endif
#endif

/// Global structure to manage states
struct sp_states g_sp_state;

/// Global packet structure for incoming and outcoming packets
struct sp_packet g_sp_packet;

/// Current bit to be sent
uint8_t g_sp_sisr_bit;
/// Current byte to be received
uint8_t g_sp_sisr_byte;
/// Pointer into the buffer to send used by the sending isr
uint8_t* gp_sp_sisr;
/// Pointer to the end of buffer 
uint8_t* gp_sp_sisr_end;

/// Obsolete
/// Lookup table to translate lap time in number of bits
/// using a 16MHz clock and a clock division by 16
// PROGMEM
//uint8_t gp_sp_timing[] = {1,1,1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,4,5,5,5,5,6,6,6,6,7,7,7,7,8,8,8,8,8};    

/// Buffer for raw data
uint8_t g_sp_buffer[SP_BUFRAWSIZE];

#if (defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL))
/// debug switches
uint16_t gb_debug = 0;
#endif

#ifdef CFG_DEBUG_PUSH
uint8_t* g_sp_debug=(uint8_t*)(SP_DEBUG_BUF);
uint16_t g_sp_debug_pos=0;
#endif

/// Synchronisation sequence used for the smartport and floppy disk
/// SP_PKTBEGIN is only for the smartport and mark the start of a packet 
const uint8_t const g_sp_sync[] PROGMEM ={0xFF,0x3F,0xCF,0xF3,0xFC,0xFF,SP_PKTBEGIN};

/// additionnal delay before sending packet to the host
//uint16_t g_sp_delay_send=400;

/**
Main task for the smartport protocol manager
*/
void sp_main()
{

    sp_main_state();

    switch(g_sp_state.status)
    {
        case SP_ENABLED:
        case SP_REQUEST:
            switch(g_sp_state.activity)
            {
                case SP_RECEIVE:
                    sp_receive_main();
                    break;
                    
                case SP_SEND:
                    sp_send_main();
                    break;
            }
            break;
        case SP_DISABLED:
     	     #if HARDWARE_V1 == 0
        	  rwts_read_position();
        	  #endif
        case SP_RESET:
            break;
    }
    
    sp_debug();
    
	// die if we trashed the memory
    if(g_sp_packet.canary!=SP_CANARY)
	{
      #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
		if(gb_debug & CFG_DEBUG_MASK_PROTO) printf_P(PSTR("SPB: packet buffer scrambled %04X\r\n"),g_sp_packet.canary);
		#endif
        die(LEDMASK_SP,LEDMASK_SP|LEDMASK_ERROR,LEDMASK_SP|LEDMASK_ERROR,1000);
	}
}

/**
Initialize the protocol interpreter at boot time
*/
void sp_init_cold()
{
 	 sp_init();
    for(uint8_t i=0;i<SP_MAX_DRIVE;i++)
    {
    	 g_sp_state.drive[i].flag = 0;
    }
 	 USB_HostState = HOST_STATE_Unattached;
}

/**
Initialize the protocol interpreter, keep the spvhd transparent for now
*/
void sp_init()
{
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug & CFG_DEBUG_MASK_PROTO) printf_P(PSTR("SPB: sp_init()\r\n"));
    #endif
        
    cli();
    
    // tell the host to wait until we are ready... -> NO!
    //SP_DO_OACKOFF;

    // RD line at high impedance until we need to talk to the host
    SP_DO_ORDDMUTE;

    // Pull down the lstrb line of the daisy chain -> NO!
    //SP_DO_OLSTRBOFF;

    // Set the counter 1 start value
    TCNT1 = 0;

    // Disable timer 1 for now
    TIMSK1 = 0;// enable with (1<<ICIE1)
    // Using the timer in normal mode
    TCCR1A = 0; // for CTC mode use (1<<WGM12);
    // Disable filter, and clock for now
    TCCR1B = 0;//(1<<CS11) (1<<CS10);

    // clearing pending interrupt
    TIFR1=(1<<OCIE1C)|(1<<ICF1);//(1<<TOIE1);
    
    sei();
    struct sp_states* p_state = &g_sp_state; 
    p_state->status = SP_UNKNOW;//sp_status_read();
    p_state->initialized = 0;
	 p_state->ctrlcode = 255;
    for(uint8_t i=0;i<SP_MAX_DRIVE;i++)
    {
    	 p_state->id[i] = 255;
    }
    p_state->drive_current = 255;
    p_state->drive_set = 0;
    // Force reload of the configuration file at each smartport reset
 	 p_state->drive_max = 0;
    p_state->block_address_next = 0xFFFFFFFF;
    //sp_init_packet(0,0,0,g_sp_buffer,SP_BUFRAWSIZE);
//    sp_receive_init();
}

/// Initialize USB communication and enumerate virtual drives
void sp_init_drives()
{
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_PROTO)	printf_P(PSTR("SPB: sp_init_drives()\r\n"));
	#endif
	
	// Start usb layer if required
	if(USB_HostState != HOST_STATE_Configured)
	{
		usb_init(0);
		// Wait at least 3 seconds to detect the usb stick
		uint16_t l_count = 3000;
		// Wait at least 3 seconds more if something is attached to the usb
		uint16_t l_delay = 7000;
		// Blinking counter
		uint16_t l_blink = 0;
		while(USB_HostState != HOST_STATE_Configured)
		{
			// if an usb device is present, allow more time for enumeration
			if(USB_HostState!=HOST_STATE_Unattached && l_delay !=0)
			{
				l_count += l_delay;
				l_delay = 0;
			}
			// blink the power led on at least 1/4 second basis max until the usb enumeration start
			if(l_blink++>500)
			{
				l_blink=0;
			}
			if(l_blink>250)
			{
				LEDs_TurnOffLEDs(LEDMASK_POWER);
			}
			else
			{
				LEDs_TurnOnLEDs(LEDMASK_POWER);
			}
			// continue usb initialiation until the usb enumeration start
			usb_init(1);
			if(l_count--==0 && USB_HostState != HOST_STATE_Configured)
			{
				// timeout, no storage available...
				LEDs_TurnOffLEDs(LEDMASK_POWER);
	/*			#if (LUFA_VERSION==100219)
				USB_ShutDown();
				#else
				USB_Disable();
				#endif
	*/			sp_init_drives_error(PSTR("No usb device"));
				return;			
			}
			_delay_ms(1);
		}
		LEDs_TurnOffLEDs(LEDMASK_POWER);
	
		// usb comunication ready with the device, try to mount the file system
		FRESULT rc = f_mount(0,&g_sp_state.filesystem);
		if(rc!=FR_OK)
		{
			#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
			if(gb_debug & CFG_DEBUG_MASK_PROTO)	printf_P(PSTR("FAT: unable to mount the file system\r\n"));
			#endif
			sp_init_drives_error(PSTR("No file system"));
			return;			
		}
		
		/* chdir don't work very well, better to avoid it
		rc = f_chdir("spvhd");
		if(rc!=FR_OK)
		{
			#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
			if(gb_debug & CFG_DEBUG_MASK_PROTO)			printf_P(PSTR("FAT: spvhd folder not found\r\n"));
			#endif
			die(LEDMASK_POWER,LEDMASK_WARNING,LEDMASK_ERROR,250);
		}
		*/
	}
	
	if(!sp_init_from_file())
	{
		// enumerate disks images to emulate drives
/*		uint8_t i,j=0;
		for(i=0;i<SP_MAX_DRIVE;i++)//-1
		{
			char filename[24];
			FILINFO fileinfo;
	
			sprintf_P(filename,PSTR("0:/spvhd/drive%d.img"),i);
			#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
			if(gb_debug & CFG_DEBUG_MASK_PROTO)	printf_P(PSTR("Looking for image %s\r\n"),filename);
			#endif
			
			// ignore missing disk image
			rc = f_stat(filename,&fileinfo);
			if(rc == FR_OK)
			{
				#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
				if(gb_debug & CFG_DEBUG_MASK_PROTO)	printf_P(PSTR("Opening image %s\r\n"),filename);
				#endif
				
				// close the file handle if already opened
				g_sp_state.drive must be setup out of sp_init()
				if(g_sp_state.drive[j]!=255)
				{
					f_close(&g_sp_state.drive[j]);
					&g_sp_state.drive[j] = 255;
				}
								
				// open the disk image and keep it opened all the time
				// ignore disk image we can't open
				rc = f_open(&g_sp_state.drive[j],filename,FA_OPEN_EXISTING | FA_READ | FA_WRITE);
				if(rc == FR_OK)
				{
					#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
					//if(gb_debug & CFG_DEBUG_MASK_PROTO)	printf_P(PSTR("Image id=%u flag=%d size=%lu cluster org=%lu\r\n"),g_sp_state.drive[i].id,g_sp_state.drive[i].flag,g_sp_state.drive[i].fsize,g_sp_state.drive[i].sclust);
					#endif
					// update the dib to reflect image size
					memcpy_P(&g_sp_state.dib[j][0],g_sp_dib_template,SP_DIBSIZE);
					g_sp_state.dib[j][0] = 0xF8;
					uint32_t size32 = (uint32_t)(fileinfo.fsize)/(uint32_t)(SP_BLOCKSIZE);
					g_sp_state.dib[j][1] = size32;
					g_sp_state.dib[j][2] = size32>>8;
					g_sp_state.dib[j][3] = size32>>16;
					// update identification string
					sprintf_P((char*)&g_sp_state.dib[j][5],PSTR("Drive%d"),i);
					g_sp_state.dib[j][4] = strlen((char*)&g_sp_state.dib[j][5]);
					g_sp_state.dib[j][g_sp_state.dib[j][4]+5]=0;
					#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
					if(gb_debug & CFG_DEBUG_MASK_PROTO)	printf_P(PSTR("Image size is %lu blocs lo %u med %u hi %u Name length %u Name !%s!\r\n"),size32,g_sp_state.dib[j][1],g_sp_state.dib[j][2],g_sp_state.dib[j][3],g_sp_state.dib[j][4],&g_sp_state.dib[j][5]);
					#endif
					g_sp_state.dib[j][g_sp_state.dib[j][4]+5]=' ';
					g_sp_state.number[j]=i;
					j++;
					g_sp_state.drive_max = j;
				}
			}
		}
*/		if(g_sp_state.drive_max==0)
		{
			sp_init_drives_error(PSTR("No disk image"));
	//die(LEDMASK_POWER,0,0,1000);                
			LEDs_SetAllLEDs(0);
		}
		else
		{
			LEDs_SetAllLEDs(LEDMASK_POWER);
		}
	}
	//g_sp_state.drive_descriptor_index = 255;
}

/// Setup the spvhd from the configuration file
uint8_t sp_init_from_file()
{
	uint8_t l_rc;
	uint8_t i;
	// recycle the end of the buffer used for nibbled data within the packet object
	uint8_t* lp_buffer = &g_sp_packet.data[SP_BUFDATSIZE - SPVHD_CFG_BLOCKSIZE]; 
	char *lp_filename = (char*)(lp_buffer - SPVHD_CFG_BLOCKSIZE);
	FILINFO* lp_fileinfo = (FILINFO*)(lp_filename - sizeof(FILINFO));
	FIL* lp_fh = (FIL*)(lp_fileinfo - sizeof(FIL));;			

	// Close all curently opened disk image
	for(i=0;i<SP_MAX_DRIVE;i++)
	{
		if(g_sp_state.drive[i].flag!=0)
		{
			f_close(&g_sp_state.drive[i]);
			g_sp_state.drive[i].flag = 0;
		}
	}
	
	g_sp_state.drive_max = 0;
	g_sp_state.drive_set = 0;
	//CSET_BANK_SET;
	strcpy_P(lp_filename,PSTR("0:/spvhd/cfgboot.dat"));
	//CSET_BANK_UNSET;
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_PROTO)		printf_P(PSTR("SPB: looking for %s\r\n"),lp_filename);
	#endif
	
	//CSET_BANK_SET;
	l_rc = f_stat(lp_filename,lp_fileinfo);
	//CSET_BANK_UNSET;
	if(l_rc != FR_OK || lp_fileinfo->fsize == 0)
	{
		#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
		if(gb_debug & CFG_DEBUG_MASK_PROTO)	printf_P(PSTR("SPB: no configuration file, creating...\r\n"));
		#endif
		// The configuration file doesn't exist, creating a new one
		l_rc = sp_cfg_file_create(lp_buffer, lp_filename, lp_fh);
	}
	if(l_rc == FR_OK)
	{
		//CSET_BANK_SET;
		l_rc = f_open(lp_fh,lp_filename,FA_OPEN_EXISTING | FA_READ);
		//CSET_BANK_UNSET;
		if(l_rc == FR_OK)
		{
			#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
			if(gb_debug & CFG_DEBUG_MASK_PROTO)	printf_P(PSTR("SPB: found config\r\n"));
			#endif
			// Read the spvhd dcb to check signature and file version
			// reuse buffer previously allocated 
			struct spvhd_dcb* lp_dcb = (struct spvhd_dcb*)(lp_buffer);
			UINT l_size_read=0;
			//CSET_BANK_SET;
			l_rc = f_read(lp_fh,(void*)lp_dcb,(WORD)SPVHD_CFG_BLOCKSIZE,&l_size_read);
			//CSET_BANK_UNSET;
			if(l_rc == FR_OK && l_size_read==SPVHD_CFG_BLOCKSIZE)
			{
				// Can read only configuration file from the same major version and with the right signature
				//CSET_BANK_SET;
				if(lp_dcb->major == SPVHD_VERSION_MAJOR && lp_dcb->signaturesize == 5 && memcmp_P(&lp_dcb->signature,PSTR("SPVHD"),5)==0 )
				{
					#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
					if(gb_debug & CFG_DEBUG_MASK_PROTO)	printf_P(PSTR("SPB: config version ok\r\n"));
					#endif
					// Ok, we have a configuration file
					
					// reuse again the same buffer					
					struct spvhd_parms* lp_parms = (struct spvhd_parms*)(lp_buffer);
					
					uint8_t l_max_drive = SP_MAX_DRIVE;
					//CSET_BANK_SET;
					if(lp_dcb->devicecount<l_max_drive)
						l_max_drive = lp_dcb->devicecount;
					//CSET_BANK_UNSET;
					uint8_t i,j=0;
					UINT l_position = 0;
					for(i=0;i<l_max_drive;i++)
					{
						#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
						if(gb_debug & CFG_DEBUG_MASK_PROTO)	printf_P(PSTR("SPB: reading virtual device %d config\r\n"),i);
						#endif
						// Loading the configuration bloc for this virtual device
						l_position += SPVHD_CFG_BLOCKSIZE;
						//l_rc = f_lseek(lp_fh,l_position);
						//if(l_rc==FR_OK)
						l_rc = f_read(lp_fh,(const void*)lp_parms,(UINT)SPVHD_CFG_BLOCKSIZE,&l_size_read);
						if(l_rc==FR_OK)
						{
							// Is the virtual drive should be enabled ?
							//CSET_BANK_SET;
							if( (lp_parms->parms&3) != 0 )
							{
								//CSET_BANK_UNSET;
								#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
								if(gb_debug & CFG_DEBUG_MASK_PROTO) printf_P(PSTR("SPB: enabling virtual device %d\r\n"),i);
								#endif
								// Enabling this virtual device
								memcpy_P(&g_sp_state.dib[j][0],g_sp_dib_template,SP_DIBSIZE);
								g_sp_state.dib[j][0] = 0xF8 | ((lp_parms->parms&3)==2?4:0);
								g_sp_state.dib[j][1] = 0;
								g_sp_state.dib[j][2] = 0;
								g_sp_state.dib[j][3] = 0;
								// update identification string
								strcpy_P((char*)&g_sp_state.dib[j][5],PSTR(""));
								g_sp_state.dib[j][4] = strlen((char*)&g_sp_state.dib[j][5]);
								g_sp_state.dib[j][g_sp_state.dib[j][4]+5]=' ';
								// Is the disk image should be mounted ?
								//CSET_BANK_SET;
								if( (lp_parms->parms&3) > 1 )
								{
									lp_parms->path[lp_parms->pathsize] = 0;
									sprintf_P(lp_filename,PSTR("0:/spvhd/%s"),lp_parms->path);
									//CSET_BANK_UNSET;
									#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
									if(gb_debug & CFG_DEBUG_MASK_PROTO) printf_P(PSTR("SPB: looking for image %s\r\n"),lp_filename);
									#endif
									
									// ignore missing disk image
									//CSET_BANK_SET;
									l_rc = f_stat(lp_filename,lp_fileinfo);
									if(l_rc == FR_OK)
									{
										//CSET_BANK_UNSET;
										#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
										if(gb_debug & CFG_DEBUG_MASK_PROTO) printf_P(PSTR("SPB: opening image\r\n"));
										#endif							
										
										// open the disk image and keep it opened all the time
										// if the disk image is read only, open it read only
										// ignore disk image we can't open
										//CSET_BANK_SET;
										l_rc = f_open(&g_sp_state.drive[j],lp_filename,FA_OPEN_EXISTING | FA_READ | ( (lp_fileinfo->fattrib&1)==0?FA_WRITE:0) );
										//CSET_BANK_UNSET;
										if(l_rc == FR_OK)
										{
											#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
											//if(gb_debug & CFG_DEBUG_MASK_PROTO)	printf_P(PSTR("Image id=%u flag=%d size=%lu cluster org=%lu\r\n"),g_sp_state.drive[i].id,g_sp_state.drive[i].flag,g_sp_state.drive[i].fsize,g_sp_state.drive[i].sclust);
											#endif
											// update the dib to reflect image size
											uint32_t l_size32 = (uint32_t)(lp_fileinfo->fsize)/(uint32_t)(SP_BLOCKSIZE);
											g_sp_state.dib[j][1] = l_size32;
											g_sp_state.dib[j][2] = l_size32>>8;
											g_sp_state.dib[j][3] = l_size32>>16;
											// update identification string
											/// @todo update the dib with the disk image file name ?
											sprintf_P((char*)&g_sp_state.dib[j][5],PSTR("Drive%d"),i);
											g_sp_state.dib[j][4] = strlen((char*)&g_sp_state.dib[j][5]);
											g_sp_state.dib[j][g_sp_state.dib[j][4]+5]=0;
											#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
											if(gb_debug & CFG_DEBUG_MASK_PROTO) printf_P(PSTR("SPB: Image size is %lu blocs lo %u med %u hi %u Name length %u Name !%s!\r\n"),l_size32,g_sp_state.dib[j][1],g_sp_state.dib[j][2],g_sp_state.dib[j][3],g_sp_state.dib[j][4],&g_sp_state.dib[j][5]);
											#endif
											g_sp_state.dib[j][g_sp_state.dib[j][4]+5]=' ';
											g_sp_state.number[j]=i;
										}
									}
									else
									{
										#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
										if(gb_debug & CFG_DEBUG_MASK_PROTO) printf_P(PSTR("SPB: missing disk image %s\r\n"),lp_filename);
										#endif
									}
									//CSET_BANK_UNSET;
									j++;
									g_sp_state.drive_max = j;
								}
								else
								{
									#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
									if(gb_debug & CFG_DEBUG_MASK_PROTO) printf_P(PSTR("SPB: virtual device kept unmounted\r\n"),lp_filename);
									#endif
								}
								//CSET_BANK_UNSET;
							}
							else
							{
								#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
								if(gb_debug & CFG_DEBUG_MASK_PROTO) printf_P(PSTR("SPB: virtual device disabled\r\n"));
								#endif
							}
							//CSET_BANK_UNSET;
						}
						else
						{
							#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
							if(gb_debug & CFG_DEBUG_MASK_PROTO) printf_P(PSTR("SPB: erreur pendant le positionnement dans la config\r\n"));
							#endif
						}
					}
					if(g_sp_state.drive_max==0)
					{
						sp_init_drives_error(PSTR("No disk image"));
						LEDs_SetAllLEDs(0);
					}
					else
					{
						LEDs_SetAllLEDs(LEDMASK_POWER);
					}
					return true;
				}
				//CSET_BANK_UNSET;
			}
			f_close(lp_fh);
		}
	}
	return false;
}

/// Create a blank configuration file
/// Use the end of g_sp_packet.data
/// @param lp_buffer Pointer to a buffer of SPVHD_CFG_BLOCKSIZE bytes
/// @param lp_filename Pointer to the filename to be created
/// @param lp_fileinfo Pointer to a file info structure
/// @param lp_fh Pointer to a file handle
/// @return FR_OK if succeded
///
uint8_t sp_cfg_file_create(uint8_t* lp_buffer,uint8_t* lp_filename, FIL* lp_fh)
{
	uint8_t l_rc;
	uint8_t i;

	l_rc = f_open(lp_fh,lp_filename,FA_CREATE_NEW | FA_WRITE);
	if(l_rc == FR_OK)
	{
		#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
		if(gb_debug & CFG_DEBUG_MASK_PROTO)			printf_P(PSTR("SPB: open...\r\n"));
		#endif
		UINT l_size_write;
		// reuse buffer previously allocated 
		struct spvhd_dcb* lp_dcb = (struct spvhd_dcb*)(lp_buffer);
		lp_dcb->major = SPVHD_VERSION_MAJOR; 
		lp_dcb->minor = SPVHD_VERSION_MINOR; 
		lp_dcb->revision = SPVHD_RESVISION;
		lp_dcb->spid = 0;
		lp_dcb->devicecount = SP_MAX_DRIVE;
		lp_dcb->signaturesize = 5;
		memcpy_P(lp_dcb->signature,PSTR("SPVHD"),5);
		l_rc = f_write(lp_fh,(const void*)lp_dcb,(UINT)SPVHD_CFG_BLOCKSIZE,&l_size_write);
		if(l_rc == FR_OK)
		{
			#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
			if(gb_debug & CFG_DEBUG_MASK_PROTO)				printf_P(PSTR("SPB: added dcb\r\n"));
			#endif
			// reuse again the same buffer					
			struct spvhd_parms* lp_parms = (struct spvhd_parms*)(lp_buffer);
			lp_parms->parms = 3;
			strcpy_P(lp_parms->path,PSTR("drive?.img"));
			lp_parms->pathsize = strlen(lp_parms->path);
			for(i=0;i<SP_MAX_DRIVE && l_rc==FR_OK;i++)
			{
				lp_parms->path[5]='0'+i;
				l_rc = f_write(lp_fh,(const void*)lp_parms,(UINT)SPVHD_CFG_BLOCKSIZE,&l_size_write);
				if(l_rc == FR_OK)
				{
					#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
					if(gb_debug & CFG_DEBUG_MASK_PROTO)						printf_P(PSTR("SPB: added config block for virtual device %d...\r\n"),i);
					#endif
				}
				else
				{
					#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
					if(gb_debug & CFG_DEBUG_MASK_PROTO)						printf_P(PSTR("SPB: write failed %02x\r\n"),l_rc);
					#endif
				}
			}
		}
		else
		{
			#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
			if(gb_debug & CFG_DEBUG_MASK_PROTO)				printf_P(PSTR("SPB: write failed %02x\r\n"),l_rc);
			#endif
		}
		f_close(lp_fh);
	}
	else
	{
		#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
		if(gb_debug & CFG_DEBUG_MASK_PROTO)			printf_P(PSTR("SPB: creation failed %02x\r\n"),l_rc);
		#endif
	}
	return l_rc;
}

/// Create a phantom virtual drive in case of major error
/// @param p_str Pointer in progmem to the name to give to the phantom drive
///
void sp_init_drives_error(const char* const p_str)
{
	uint8_t j = 0;
	memcpy_P(&g_sp_state.dib[j][0],g_sp_dib_template,SP_DIBSIZE);
	g_sp_state.dib[j][0] = 0xF8;
	g_sp_state.dib[j][1] = 0;
	g_sp_state.dib[j][2] = 0;
	g_sp_state.dib[j][3] = 0;
	// update identification string
	sprintf_P((char*)&g_sp_state.dib[j][5],p_str);
	g_sp_state.dib[j][4] = strlen((char*)&g_sp_state.dib[j][5]);
	g_sp_state.dib[j][g_sp_state.dib[j][4]+5]=' ';
	g_sp_state.drive_max = 1;
	//LEDs_SetAllLEDs(LEDMASK_ERROR);
}

/**
Initialize the sending and receving of a packet
@param mode Set to 0 to receive data, set to 1 to send data
@param type Type of the packet to write in the header
@param status Status byte to write in the header
@param p_buffer POinter to the data buffer
@param size Size of the data buffer
*/
void sp_init_packet(uint8_t mode,uint8_t type,uint8_t status,uint8_t* p_buffer,uint16_t size)
{
	struct sp_states* p_state = &g_sp_state; 
	if(mode==0)
	{
		// receive data
		// waiting for something to do
		p_state->p_packet = (uint8_t*)&g_sp_packet.header;
		p_state->p_packet_end = (uint8_t*)(&g_sp_packet.canary);
		p_state->activity = SP_RECEIVE;
		p_state->receive_step = SP_WAIT;		
	}
	else
	{
		// send data
		// sync sequence required only in send mode
		memcpy_P(g_sp_packet.sync,g_sp_sync,sizeof(g_sp_sync));
		g_sp_packet.header.iddst = 128;
		uint8_t drive_current = p_state->drive_current;
		if(drive_current < p_state->drive_set)
			g_sp_packet.header.idsrc = p_state->id[drive_current]|128;
		else
			g_sp_packet.header.idsrc = 128;
		g_sp_packet.header.pkttype = type|128;
		g_sp_packet.header.pktauxtype = 128;
		g_sp_packet.header.status = status|128;
		p_state->activity = SP_SEND;
		p_state->send_step = SP_WAIT;

    /*
;g_sp_sisr_bit is the current bit to be sent and must be set to 0x80
;g_sp_sisr_byte must be set with the first byte to send
;gp_sp_sisr must be set to the start of the buffer to send +1
;gp_sp_sisr_end must be set to the end of the buffer to send +1
    */
		gp_sp_sisr = (uint8_t*)(&g_sp_packet.sync[1]);
		g_sp_sisr_bit = 0x80;
		g_sp_sisr_byte = g_sp_packet.sync[0];
	}
	p_state->p_buffer = p_buffer;
	p_state->p_buffer_end = &p_buffer[size];
	p_state->size_computed = size;
	p_state->error = SP_ERROK;
	g_sp_packet.canary = SP_CANARY;
}  

/**
Convert a size to the smartport format
@param sp_size Pointer to the sp_size structure to set
@param size The size to set to
@return void
*/
void sp_size_set(struct sp_size* p_obj,uint32_t size)
{
  p_obj->low=size;
  p_obj->middle=size>>8;
  p_obj->high=size>>16;
}

/**
Convert a size from the smartport format
@param sp_size Pointer to the sp_size structure to read
@return The size got from the sp_size structure
*/
uint32_t sp_size_get(struct sp_size* p_obj)
{
  return ((uint32_t)(p_obj->low))|(((uint32_t)(p_obj->middle))<<8)|(((uint32_t)(p_obj->high))<<16);
}

/**
Retreive the lowest smartport id managed by the spvhd
@return The lowest smartport id managed by the spvhd, or 255 if not found
*/
uint8_t sp_lowest_id_get(struct sp_states* p_state)
{
	for(uint8_t i=0;i<SP_MAX_DRIVE;i++)
	{
		if(p_state->id[i]!=255)
		{
			return p_state->id[i];
		}
	}
	return 255;
}

