/*
=============================================================================
Smartport Virtual Hard Drive, version 1.00.00
smartport (aka Cbus) protocol interpreter

If you would like to be involved in this project, or simply get my 
documentation for your own smartport project, feel free to contact me at
pcedric@hotmail.com

Copyright 2010,2011,2012 Cédric PELTIER, pcedric@hotmail.com

This file is part of the Smartport VHD project

Smartport VHD is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 3

Smartport VHD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Smartport VHD. If not, see <http://www.gnu.org/licenses/>.
=============================================================================
*/

/**
Some explanations

The library is composed of two layers running together:
Layer 1: manage the protocol and respond to the requests sent by the host
- protocol.c
Layer 0: manage basic signaling as well as sending and receiving packets
- protocol_low.c protocol_rcv_isr.c protocol_snd_isr.c

Received byte are stored directly in the packet buffer by the layer 0. 
The first approach was to reciev and store incoming data by the layer 0 and let
the layer 1 decode the data and stop the layer 0 at the end of the packet.
A 'canary' word at the end of the packet buffer prevented to scramble the memory
if the upper layer 1 don't stoped the layer 0 in time.
But this way was too slow to acknowledge packets in time, so it was discarded
and the data packet structure is decoded on the fly within the layer 0.

The differents aproachs to decode incoming data from the host are exposed here:

Method 1: 
Each msb is detected by a PCINT pin, raising isr1 that activate the sampling of
the WR line by isr2 and desactivate itself.
The isr2 is raised 7 times each 4us to store the byte, then activate again isr1
for the next msb, change his time limit to serve as a watchdog for isr1 for the
last byte, and store the new byte directly into the packet buffer.
This method raise one interrupt per bit, and the average consumed cpu power is
around 37%. Another isr should not run concurently as it will introduce
unpredictables delays.

Method 2:
Each toggle on the WR line is detected by the hardware on the ICP1 pin, raising
an interrupt with a timestamp capture. The interrupt routine compute how many
bits set to 0 are in the lap time. Each receveid byte goes directly into the
packet buffer. After the first received msb, a watchdog is set to ensure the
storage of the last byte. Using the hardware timestamp capture give accurate
measure. 
This method take an average of 1.96 bits per interrupt, wich should help to lower
the isr overhead, but the required 16 bit arithmetic push the cpu power 
consumption up to 44%...

Method 3:
From the method 2, we can reduce interrupt contention by moving time delta
computing from the isr to the upper layer. This require a quite huge buffer
to store timestamps, but the isr is simpler, and the overall computing is
more efficient as register usage is lowered significantly in the isr.

Method 4:
Use the brute force of the avr to sample continuously the WR line, consuming
100% of the cpu power, or requiring high integration of the layers and the
application.

Method 5, selected:
Each toggle on the WR line is detected by the hardware on the ICP1 pin, raising
an interrupt. The isr wait for each msb and sample the WR line on a 4us basis
using hardware timestamp. This is a brute force method like method 4. Each
received byte are interpreted on the fly. Protocol's bytes are stored into the
packed structure, and packet data bytes are stored directly inside the receive 
buffer. The crc is computed on the fly too, so we can acknowledge the packet in 
time. 

The problem with small isr that require to be raised often is that each register
used by the isr must be saved and restored once the isr end. To speedup things 
up we can reserve some register so the C compiler will not use them, but the 
libc will still use these registers so it's of no help.

I prefer small isr that does only what is required, but as the data have to be
decoded on the fly, and the isr overhead is too big to capture each bits, the
reception of one packet is done directly inside the isr, wich become a monster.

The receiving isr is wrote in C and use a lot of registers, it's called once per
packet and keep the control until the end of the packet.

The sending isr is wrote in assembly and use as few registers as possible; it's
called once per bit to send until it reach the end of the buffer to be sent.
PWM mode is used to generate negatives pulses of 1us width on a 4us basis.
When a bit set to 0 have to be sent, the output of the PWM pin is temporarily 
disconnected by the sending isr, so the pulse is discarded but the timing is 
kept. This isr is started from sp_encode() once the sync and the header bytes 
are ready. Doing so we are encoding and sending the packet at the same time.
It's ok for the smartport slow mode at 4us per bit, but unusable with the fast 
mode at 2us per bit, as the isr overhead is too big.

OCF1C is set when TCNT1 match OCR1C
OCF1C is cleared by writing 1 into OCF1C of TIFR1


*/


#ifndef _PROTOCOL_H_
#define _PROTOCOL_H_

    #ifndef __ASSEMBLER__
    #include <stdint.h>
    #include "Lib/FATFs/ff.h"
    #endif

// =============================================================================
// dedicated registers

	#ifdef CFG_SP_ISR_USE_DEDICATED
    #ifdef __ASSEMBLER__
    
        /// registers used by assembly isr

        /// saving sreg here
        //#define g_sp_isr_sreg r2
        
        /// temporary register
        //#define g_sp_isr_r3 r3

        /// doh...X is dedicated !!!
        //#define g_sp_isr_r26 r26
        //#define g_sp_isr_r27 r27

        /// saving the last WR line state here to compare
        //#define g_sp_isr_state r17

        /// temporary register for isr
        //#define g_sp_isr_tmp_h r19
        //#define g_sp_isr_tmp_l r18

        /// current event timestamp
        //#define g_sp_isr_new_h r5
        //#define g_sp_isr_new_l r4

        /// Pointer to write within the ring buffer
        //#define gp_sp_isr_h r7
        //#define gp_sp_isr r6

        /// bitmask used to change the edge detector
        //#define g_sp_isr_bit r3

    #else
    
        /// registers dedicated to assembly isr
        /// disabled as this require to recompile all libraries

        //register uint8_t g_sp_isr_sreg asm("r2");
        //register uint8_t g_sp_isr_r3 asm("r3");
        //register uint8_t g_sp_isr_r26 asm("r26");
        //register uint8_t g_sp_isr_r27 asm("r27");
        //register uint8_t g_sp_isr_state asm("r17");

        /// temporay pointer and 16bit register for isr
        //register uint16_t g_sp_isr_tmp_l asm("r18");

        /// current event timestamp
        //register uint16_t g_sp_isr_new_l asm("r4");

        /// Pointer to read/write for isr
        //register uint8_t* gp_sp_isr asm("r6");

        /// bitmask used to change the edge detector
        //register uint8_t g_sp_isr_bit asm("r3");

    #endif
    #endif

// =============================================================================
// Definitions for software

    /// packet bytes
    #define SP_PKTSYNC 0xFF    // byte used to synchronize before packet header
    #define SP_PKTBEGIN 0xC3   // start of a packet
    #define SP_PKTEND  0xC8    // end of a packet

    #define SP_PKTCMD  0x00    // command packet
    #define SP_PKTSTA  0x01    // status packet
    #define SP_PKTDAT  0x02    // data packet

    /// command bytes
    #define SP_CMDSTATUS 0x00    // status request
    #define SP_CMDREAD   0x01    // read sector request
    #define SP_CMDWRITE  0x02    // write sector request
    #define SP_CMDFORMAT 0x03    // format media request
    #define SP_CMDCTRL   0x04    // control request
    #define SP_CMDSID    0x05    // assign id request

    /// status bytes
    #define SP_STASTA 0x00    // device status
    #define SP_STADCB 0x01    // device control block
    #define SP_STADIB 0x03    // device information block

    /// control bytes
    #define SP_CTRLRST 0x00    // device soft reset
    #define SP_CTRLDCB 0x01    // set device control block
    #define SP_CTRLNLS 0x03    // device information block
    #define SP_CTRLINT 0x04    // service device interrupt

    /// smartport error code
    #define SP_ERR_OK		0x00	// no error
    #define SP_ERR_BADCMD	0x01	// invalid command
    #define SP_ERR_BUSERR	0x06	// communication error
    #define SP_ERR_BADCTL	0x21	// invalid control code
    #define SP_ERR_BADPRM	0x22	// invalid parameter list
    #define SP_ERR_IOERROR	0x27	// I/O error
    #define SP_ERR_NODRIVE	0x28	// no device connected
    #define SP_ERR_NOWRITE	0x2B	// disk write protected
    #define SP_ERR_BADBLOCK 0x2D // invalid block number
    #define SP_ERR_DSKSWITCHED 0x2E // disk switched, file open on another disk
    #define SP_ERR_OFFLINE	0x2F	// device offline of no disk in drive
    
    /// buffer size for nibbled data within a packet
    #define SP_BUFDATSIZE 590

    /// buffer size for raw data (used only for command packets)
    #define SP_BUFRAWSIZE 64
    //512 (full block size)
    //+2 (16 bits canary)

    /// size of a block
    #define SP_BLOCKSIZE 512

    /// size of the device bloc information
    #define SP_DIBSIZE 25
    
    /// debug ring buffer
    #define SP_DEBUG_BUF  0x8000
    #define SP_DEBUG_BUF_SIZE 16384

    /// canary value to check buffer overflow
    #define SP_CANARY   0x4132
    #define SP_CANARYH  0x41
    #define SP_CANARYL  0x32
    
// =============================================================================
// Definitions for hardware

    // lines mapping between smartport bus and us
    
    /// PORT B outputs
    /// B0 /INT		to chained device
    /// B1 				reserved spi
    /// B2 				reserved spi
    /// B3 				reserved spi
    /// B4 ACK			to bus
    /// B5 CA3 LSTRB to chained device
    /// B6 /DR2		to chained device
    /// B7 RD			to bus
    
    /// PORT D inputs
    /// D0 				reserved i2c
    /// D1 				reserved i2c
    /// D2 				reserved uart
    /// D3 				reserved uart
    /// D4 WR			from bus
    /// D5 /DR2		from host
    /// D6 HDSEL 		from chained device
    /// D7 35SEL 		from bus
    
    /// PORT F inputs
    /// F0 PH0 IREQ 	from bus
    /// F1 PH1 CA1 	from bus
    /// F2 PH2 CA2 	from bus
    /// F3 PH3 LSTRB from host
    /// F4 /INT 		from host
	 /// F5 WREQ		from bus
	 /// F6 free
    /// F7 free
    
    /// input from the host, request line
    #define SP_IREQ_PORT PORTF
    #define SP_IREQ_PIN PINF
    #define SP_IREQ_BN PF0
    #define SP_IREQ    (1<<SP_IREQ_BN)
    
    /// input from the host, phase 1 line
    #define SP_ICA1_PORT PORTF
    #define SP_ICA1_PIN PINF
    #define SP_ICA1_BN PF1
    #define SP_ICA1    (1<<SP_ICA1_BN)
    
    /// input from the host, phase 2 line
    #define SP_ICA2_PORT PORTF
    #define SP_ICA2_PIN PINF
    #define SP_ICA2_BN PF2
    #define SP_ICA2    (1<<SP_ICA2_BN)
    
    /// input from the host, phase 3 line
    #define SP_ILSTRB_PORT PORTF
    #define SP_ILSTRB_PIN PINF
    #define SP_ILSTRB_BN PF3
    #define SP_ILSTRB (1<<SP_ILSTRB_BN)
    
    /// input from the host, apple 3.5 select
    #define SP_I35SEL_PORT PORTD
    #define SP_I35SEL_PIN PIND
    #define SP_I35SEL_BN PD7
    #define SP_I35SEL (1<<SP_I35SEL_BN)

    /// input from the daisy chain to detect the presence of intelligent resident
    #define SP_IHDSEL_PORT PORTD
    #define SP_IHDSEL_PIN PIND
    #define SP_IHDSEL_BN PD6
    #define SP_IHDSEL (1<<SP_IHDSEL_BN)

    /// output to the chained device, phase 3
    #define SP_OLSTRB_PORT PORTB
    #define SP_OLSTRB_PIN PINB
    #define SP_OLSTRB_BN PB5
    #define SP_OLSTRB  (1<<SP_OLSTRB_BN)
    
    /// input from the bus, write request
    #define SP_IWREQ_PORT PORTF
    #define SP_IWREQ_PIN PINF
    #define SP_IWREQ_BN PF5
    #define SP_IWREQ  (1<<SP_IWREQ_BN)

    /// input from the host, 1st external 5.25 enabled
    #define SP_IDR2_PORT PORTD
    #define SP_IDR2_PIN PIND
    #define SP_IDR2_BN PD5
    #define SP_IDR2    (1<<SP_IDR2_BN)

    /// ouput to the chained device, 1st external 5.25 enabled
    /// kept high
    /// - until all virtual devices are initialized with their ids
    /// - during smartport resident communications
    /// - the drive is virtualized
    /// copy of IDR2 otherwise
    #define SP_ODR2_PORT PORTB
    #define SP_ODR2_PIN PINB
    #define SP_ODR2_BN PB6
    #define SP_ODR2    (1<<SP_ODR2_BN)

    /// input from the host, 2nd external 5.25 enabled
    #define SP_IINT_PORT PORTF
    #define SP_IINT_PIN PINF
    #define SP_IINT_BN PF4
    #define SP_IINT    (1<<SP_IINT_BN)

    /// output to the chained device, 2nd external 5.25 enabled
    /// kept high
    /// - until all virtual devices are initialized with their ids
    /// - during smartport resident communications
    /// - the drive is virtualized
    /// copy of IINT otherwise
    #define SP_OINT_DDR DDRB
	 #define SP_OINT_PORT PORTB
    #define SP_OINT_PIN PINB
    #define SP_OINT_BN PB0
    #define SP_OINT    (1<<SP_OINT_BN)

    /// output to the host, read data
    /// high impedance until the device must send data
    /// output during communications and grounded 4us for each bit set to 1
    #define SP_ORDD_DDR DDRB
    #define SP_ORDD_PORT PORTB
    #define SP_ORDD_PIN PINB
    #define SP_ORDD_BN PB7
    #define SP_ORDD    (1<<SP_ORDD_BN)
    
    /// input from the host, write data
    #define SP_IWRD_PORT PORTD
    #define SP_IWRD_PIN PIND
    #define SP_IWRD_BN PD4
    #define SP_IWRD    (1<<SP_IWRD_BN)

    /// output to the host, acknowledge changed from PD0 to PB4 to allow I2C RTC
    /// high impedance until the device must acknowledge data
    /// output and grounded during acknowledge
    #if HARDWARE_V1 == 1
    #define SP_OACK_DDR DDRD
    #define SP_OACK_PORT PORTD
    #define SP_OACK_PIN PIND
    #define SP_OACK_BN PD0
	 #else
    #define SP_OACK_DDR DDRB
    #define SP_OACK_PORT PORTB
    #define SP_OACK_PIN PINB
    #define SP_OACK_BN PB4
	 #endif
    #define SP_OACK    (1<<SP_OACK_BN)
    
    /// ports direction mask
    /// SP_ORDD not included in SP_OPORTB, dynamically controled and high impedance by default
    /// SP_OACK not included in SP_OPORTB, dynamically controled and high impedance by default
    /// SP_OINT not included in SP_OPORTB, dynamically controled and high impedance by default
    #define SP_OPORTB (SP_ODR2|SP_OLSTRB)
    #define SP_OPORTD 0
    #define SP_OPORTE 0
    #define SP_OPORTF 0
    
    #ifdef __ASSEMBLER__

        /// error status
        #define SP_ERROK 0
        #define SP_ERRCRC 255
        #define SP_ERRTIMEOUT 254
        #define SP_ERROVERFLOW 253
        #define SP_ERRPACKET 252
        #define SP_ERRRESET 251
        #define SP_ERREND 250
    
    #else

        /// disable ouput on ack line and let the pullup do it's job DDR=0 PORT=0
        #define SP_DO_OACKON SP_DO_OACKMUTE;SP_DEBUG_PUSH('A');

        /// pull low the ack line and enable output PORT=0 DDR=1
        #define SP_DO_OACKOFF SP_OACK_PORT &= ~SP_OACK; SP_OACK_DDR |= SP_OACK; SP_DEBUG_PUSH('a');

        /// enable the ack line DDR=1 PORT=0
        #define SP_DO_OACKENBL SP_OACK_PORT &= ~SP_OACK; SP_OACK_DDR |= SP_OACK; 

        /// mute the ack line (ack going high via the host pullup) DDR=0 and PORT=0
        #define SP_DO_OACKMUTE SP_OACK_PORT &= ~SP_OACK; SP_OACK_DDR &= ~SP_OACK;

        /// push high the dr2 line to the chained device (first daisy chained 5.25 disabled)
        #define SP_DO_ODR2ON   SP_ODR2_PORT |= SP_ODR2;
        /// push low the dr2 line to the chained device (first daisy chained 5.25 enabled)
        #define SP_DO_ODR2OFF  SP_ODR2_PORT &= ~SP_ODR2;
        // an attempt to use the duodisk on the IIgs
        //#define SP_DO_ODR2OFF if(g_sp_state.initialized!=0) {SP_ODR2_PORT &= ~SP_ODR2;}

        /// push high the int line to the chained device (2nd daisy chained 5.25 disabled)
        /// PORT=1 and DDR=0, so input with pullup in case a daisy chained device
        /// want to send an interrupt to the host, wich shouldn't arise but who know
        //#define SP_DO_OINTON   SP_OINT_PORT |= SP_OINT;
        #define SP_DO_OINTON SP_OINT_DDR &= ~SP_OINT;SP_OINT_PORT |= SP_OINT; 

        /// push low the int line to the chained device (2nd daisy chained 5.25 enabled)
        /// PORT=0 and DDR=1, so ouput grounded to select the device
        //#define SP_DO_OINTOFF  SP_OINT_PORT &= ~SP_OINT;
        #define SP_DO_OINTOFF SP_OINT_PORT &= ~SP_OINT;SP_OINT_DDR |= SP_OINT;
        // an attempt to use the duodisk on the IIgs
        //#define SP_DO_OINTOFF if(g_sp_state.initialized!=0) {SP_OINT_PORT &= ~SP_OINT;SP_OINT_DDR |= SP_OINT;} 

        /// enable the rd line to send data to the host, the rd line is set to low level PORT=0 DDR=1 (work but inverted signal)     
        //#define SP_DO_ORDDENBL SP_ORDD_PORT &= ~SP_ORDD; SP_ORDD_DDR |= SP_ORDD;
        /// enable the rd line to send data to the host, the rd line is set to high level PORT=1 DDR=1
        #define SP_DO_ORDDENBL SP_ORDD_PORT |= SP_ORDD; SP_ORDD_DDR |= SP_ORDD;

        /// mute the rd line by putting high impedance, PORT=0 DDR=0
        #define SP_DO_ORDDMUTE SP_ORDD_PORT &= ~SP_ORDD; SP_ORDD_DDR &= ~SP_ORDD; 

        /// push high the olstrb line to the chained device
        #define SP_DO_OLSTRBON  SP_OLSTRB_PORT |= SP_OLSTRB;

        /// pull low the olstrb line to the chained device
        #define SP_DO_OLSTRBOFF SP_OLSTRB_PORT &= ~SP_OLSTRB; 

        /// true if the ack line from the bus is pushed high
        #define SP_IF_OACK    ((SP_OACK_PIN & SP_OACK)==SP_OACK)

        /// true if the req line from the bus is pushed high
        #define SP_IF_IREQ    ((SP_IREQ_PIN & SP_IREQ)==SP_IREQ)

        /// true if the drive 2 line from the host is high (1st external 5.25 not selected)
        #define SP_IF_IDR2    ((SP_IDR2_PIN & SP_IDR2)==SP_IDR2)
        
        /// true if the drive 2 line is high to the daisy chain
        #define SP_IF_ODR2    ((SP_ODR2_PIN & SP_ODR2)==SP_ODR2)

        /// true if the int line from the host is high (2nd external 5.25 not selected)
        #define SP_IF_IINT    ((SP_IINT_PIN & SP_IINT)==SP_IINT)
        
        /// true if the int line is high to the daisy chain
        #define SP_IF_OINT    ((SP_OINT_PIN & SP_OINT)==SP_OINT)

        /// true if the lstrb line from the host is pushed high
        #define SP_IF_ILSTRB  ((SP_ILSTRB_PIN & SP_ILSTRB)==SP_ILSTRB)

        /// true if the 35sel line from the bus is pushed high
        #define SP_IF_I35SEL  ((SP_I35SEL_PIN & SP_I35SEL)==SP_I35SEL)
        
        /// true if the hdsel line from the daisy chain device is pushed high (no more intelligent resident)
        #define SP_IF_IHDSEL  ((SP_IHDSEL_PIN & SP_IHDSEL)==SP_IHDSEL)

        /// true if the ca2 line from the host is pushed high
        #define SP_IF_ICA2    ((SP_ICA2_PIN & SP_ICA2)==SP_ICA2)

        /// true if the ca1 line from the bus is pushed high
        #define SP_IF_ICA1    ((SP_ICA1_PIN & SP_ICA1)==SP_ICA1)

        /// true if the req line from the bus is pushed high
        #define SP_IF_IREQ    ((SP_IREQ_PIN & SP_IREQ)==SP_IREQ)

        /// true if the wreq line from the bus is pushed high
        #define SP_IF_WREQ    ((SP_WREQ_PIN & SP_WREQ)==SP_WREQ)
        
        /// true if the smartport is in reset state
        #define SP_IF_RESET   ((PINF & (SP_IREQ|SP_ICA1|SP_ICA2|SP_ILSTRB)) == (SP_IREQ|SP_ICA2))

        /// true if the smartport is enabled (/DR2 is high, else the computer is accessing external Apple 3.5 or 5.25 floppy drive)
        #define SP_IF_ENABLED (((PINF & (SP_ICA1|SP_ICA2|SP_ILSTRB)) == (SP_ICA1|SP_ILSTRB)) /*&& SP_IF_IDR2*/)

        /// true if the smartport is requesting
        #define SP_IF_REQUEST ((PINF & (SP_IREQ|SP_ICA1|SP_ICA2|SP_ILSTRB)) == (SP_IREQ|SP_ICA1|SP_ILSTRB))

		#ifdef CFG_DEBUG_PUSH
        #define SP_DEBUG_PUSH(c) sp_debug_push(c)
        #else
        #define SP_DEBUG_PUSH(c)
		#endif
        
// =============================================================================
// Structures and enumerations

        /// communication states
        enum sp_communication
        {
            SP_NONE=0,
            SP_RECEIVE=1,
            SP_SEND=2
        };

        /// activity states
        enum sp_activity
        {
            SP_RECEIVE_ID=0,
            SP_WAIT_REQUEST=1,
            SP_SEND_ACK=2
        };

        /// smartport states
        enum sp_status
        {
            SP_RESET=0,
            SP_DISABLED=1,
            SP_ENABLED=2,
            SP_REQUEST=3,
            SP_UNKNOW=4
        };
        
        /// reading states
        enum sp_step
        {
            SP_STOP=0,
            SP_WAIT=1,
            SP_SYNC=2,
            SP_HEAD=3,
            SP_DATA_ODD_MSB=4,
            SP_DATA_ODD=5,
            SP_DATA_GRP_MSB=6,
            SP_DATA_GRP=7,
            SP_CRC=8,
            SP_END=9,
            SP_GOT=10
        };

        /// error status
        enum sp_error
        {
            SP_ERROK=0,             // no error, good !
            SP_ERRCRC=1,
            SP_ERRTIMEOUT=2,
            SP_ERROVERFLOW=3,
            SP_ERRPACKET=4,       // error in the packet format
            SP_ERRRESET=5,        // smartbus reset
            SP_ERRTIMING=6        // low level error related to timings
        };
        
        typedef union
        {
            uint8_t* pointer;
            struct {
                uint8_t lo;
                uint8_t hi;
            };
        } pointer;
        
        /// structure to manage the WR line
        struct sp_receive_isr_low
        {
        	  uint8_t line;
        	  uint8_t last;
        	  uint8_t bit;
        	  uint8_t byte;
        };
        
        /// packet manipulation state structure
        struct sp_states
        {
            /// current smartport state (enabled, disabled, reset)
            enum sp_status status:3;
            /// current error state
            enum sp_error error:3;
            /// current activity (none, receive, send)
            enum sp_communication activity:2;
            /// current state for a receiving process
            enum sp_step receive_step:4;
            /// current state for a sending process
            enum sp_step send_step:4;
				/// Current virtual device for configuration access
				uint8_t devnum:7;
            /// flag indicating if all virtual drives are initialized (1) or not (0)
            uint8_t initialized:1;
            /// selector indicating how to interpret the incoming data of the data packet 
            uint8_t ctrlcode;
				/// Global pointer to write into the g_sp_packet structure
				uint8_t* p_packet;
				/// Global pointer on the end of the packet's buffer
				uint8_t* p_packet_end;
            /// start of the raw data buffer
            uint8_t* p_buffer;
            /// end of the raw data buffer
            uint8_t* p_buffer_end;
            /// computed size of raw data
            uint16_t size_computed;
            /// pin monitoring when receiving data in isr
            /// *
				uint8_t receive_line;
				/// descriptor of the current opened image file
				//FIL drive_descriptor;
				/// current drive number related to the current opened image file
				//uint8_t drive_descriptor_index;
				/// current drive index being accessed
				uint8_t drive_current;
				/// number of drives with id
				uint8_t drive_set;
				/// number of emulated drives
				uint8_t drive_max;
				/// Address of the current block
				uint32_t block_address;
				/// Address of the next block to read
				uint32_t block_address_next;
				/// File system object management
				FATFS filesystem;
				/// File object management
				FIL drive[SP_MAX_DRIVE];
				/// Our ids on the smartport bus
				uint8_t id[SP_MAX_DRIVE];
				/// Our absolute number
				uint8_t number[SP_MAX_DRIVE];
				/// Devices identifications
				/// @todo create the dib on demand
				uint8_t dib[SP_MAX_DRIVE][SP_DIBSIZE];
				/// to parse repository
				DIR directory;
        };

        /// size format for the smartport protocol
        struct sp_size
        {
            uint8_t low;
            uint8_t middle;
            uint8_t high;
        };

        /// device information block
        struct sp_dib
        {
            uint8_t dsb; 
            struct sp_size size;
            uint8_t idstringsize;
            uint8_t idstring[16];
            uint8_t devicetype;
            uint8_t devicesubtype;
            uint16_t version;
        };
        
        /// packet header structure
        struct sp_header
        {
            uint8_t iddst;     // destination
            uint8_t idsrc;     // source
            uint8_t pkttype;   // type
            uint8_t pktauxtype;// auxiliary type
            uint8_t status;    // status
            uint8_t nbodd;     // number of odd bytes
            uint8_t nbgrp;     // number of nibbles
        };
        
        /// packet structure
        struct sp_packet
        {
            /// sync sequence
            uint8_t sync[6];
            /// begin packet mark
            uint8_t begin_mark;
            /// header of the packet
            struct sp_header header;
            /// nibbled data of the packet
            uint8_t data[SP_BUFDATSIZE];
            /// overflow checkup
            uint16_t canary;
        };       
        
// =============================================================================
// Function Prototypes

        void sp_main(void);
        void test(void);
        void sp_main_state(void);
        void sp_state_enabled(void) __attribute__((always_inline));
        void sp_state_request(void) __attribute__((always_inline));
        void sp_state_disabled(void) __attribute__((always_inline));
        void sp_state_reset(void) __attribute__((always_inline));
              
        void sp_init_drives(void);
        void sp_init_drives_error(const char* const p_str);
        uint8_t sp_init_from_file(void);
        
        void sp_receive_init(void);
        void sp_receive_start(void) __attribute__((always_inline));
        void sp_receive_stop(void) __attribute__((always_inline));
        void sp_receive_main(void);
        void sp_receive_isr_end(void); // __attribute__((always_inline));
        uint8_t sp_receive_isr_get_byte(struct sp_receive_isr_low*) __attribute__((always_inline));
        void sp_receive_isr_wait_msb(struct sp_receive_isr_low*) __attribute__((always_inline));
        void sp_receive_isr_next_bit(struct sp_receive_isr_low*) __attribute__((always_inline));
        uint8_t sp_receive_isr_last_bit(struct sp_receive_isr_low*) __attribute__((always_inline));
        uint8_t sp_get_byte(void) __attribute__((always_inline)); ///*
        void sp_receive_packet(void) __attribute__((always_inline));
        uint8_t sp_receive_id(void);// __attribute__((always_inline));
        void sp_receive_command(struct sp_states* p_state);// __attribute__((always_inline));
        uint8_t sp_receive_read(void);// __attribute__((always_inline));
        uint8_t sp_receive_write(void);// __attribute__((always_inline));
        uint8_t sp_receive_status(struct sp_states* p_state);// __attribute__((always_inline));
        uint8_t sp_receive_control(struct sp_states* p_state);
        uint8_t sp_receive_dcb(struct sp_states* p_state,uint8_t ctrlcode);
        void sp_receive_data(struct sp_states* p_state);// __attribute__((always_inline));
        void sp_receive_data_block(struct sp_states* p_state);// __attribute__((always_inline));
        //void sp_receive_data_dcb(struct sp_states* p_state);// __attribute__((always_inline));
        void sp_receive_data_date(struct sp_states* p_state);
        void sp_receive_data_explore(struct sp_states* p_state);
        void sp_receive_data_vdsel(struct sp_states* p_state);
        void sp_receive_data_parm(struct sp_states* p_state);
        
        uint8_t sp_handshake_closing(void) __attribute__((always_inline));
        uint8_t sp_handshake_opening(void) __attribute__((always_inline));
        void sp_update_olstrb(void) __attribute__((always_inline));
        void sp_update_odr2(void) __attribute__((always_inline));
        void sp_update_oint(void) __attribute__((always_inline));
        
        void sp_send_main(void);
        void sp_send_init(uint8_t type,uint8_t status,uint8_t* p_buffer,uint16_t size);
        enum sp_error sp_encode(void);
        void sp_send_start(void) __attribute__((always_inline));
        void sp_send_ds(uint8_t status);// __attribute__((always_inline));
        void sp_send_dib(uint8_t status);// __attribute__((always_inline));
        void sp_send_status(uint8_t status);// __attribute__((always_inline));
        void sp_send_data(uint8_t* data);// __attribute__((always_inline));
        void sp_send_dcb(struct sp_states* p_state);// __attribute__((always_inline));
        void sp_send_date(struct sp_states* p_state);
        void sp_send_nextfile(struct sp_states* p_state);
        void sp_send_parm(struct sp_states* p_state, uint8_t l_type);

        uint8_t sp_cfg_file_create(uint8_t* lp_buffer,uint8_t* lp_filename, FIL* lp_fh);

        void sp_init(void);
        void sp_init_cold(void);
        void sp_init_packet(uint8_t mode,uint8_t type,uint8_t status,uint8_t* p_buffer,uint16_t size);
        void sp_cancel_admin(struct sp_states* p_state);
        enum sp_status sp_status_read(void) __attribute__((always_inline));
        
        void sp_size_set(struct sp_size* p_obj,uint32_t size) __attribute__((always_inline));
        uint32_t sp_size_get(struct sp_size* p_obj) __attribute__((always_inline));
        //void sp_select_image(struct sp_states* p_state) __attribute__((always_inline));
        uint8_t sp_lowest_id_get(struct sp_states* p_state);
        
        #ifdef CFG_DEBUG_PUSH
        void sp_debug_lines(void) __attribute__((always_inline));
        void sp_debug_push(uint8_t chr) __attribute__((always_inline));
        #endif
        void sp_debug(void);
        //void sp_debug_ring(void);
        void sp_debug_packet(void);

        /// Global structure to manage states
		extern struct sp_states g_sp_state;
		
		/// Global packet structure for incoming and outcoming packets
        extern struct sp_packet g_sp_packet;
        
        /// Global packet structure for incoming and outcoming packets
        extern struct sp_packet g_sp_packet;

		/// Current bit to be sent
		extern uint8_t g_sp_sisr_bit;
		/// Current byte to be sent
		extern uint8_t g_sp_sisr_byte;
		/// Pointer into the buffer to send used by the sending isr
		extern uint8_t* gp_sp_sisr;
		/// Pointer to the end of buffer 
		extern uint8_t* gp_sp_sisr_end;

        /// Lookup table to translate lap time in number of bits
        /// using a 16MHz clock and a division by 16
        // PROGMEM
        //extern uint8_t gp_sp_timing[];

        /// Buffer for unpacked data
        extern uint8_t g_sp_buffer[SP_BUFRAWSIZE];

        #if (defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL))
        /// debug switches
        extern uint16_t gb_debug;
        #endif
        
        #ifdef CFG_DEBUG_PUSH
        ///
        extern uint8_t* g_sp_debug;
        
        ///
        extern uint16_t g_sp_debug_pos;
        #endif
        
        /// device identification for the smartport
        const uint8_t const g_sp_dib_template[SP_DIBSIZE];

        /// synchronisation sequence for the smartport
        extern const uint8_t const g_sp_sync[];

        /// additionnal delay before sending packet to the host
        //uint16_t g_sp_delay_send;

        /// current drive number being accessed
        //uint8_t g_sp_drive_current;

        /// number of emulated drives
        //uint8_t g_sp_drive_max;
    #endif
        
#endif

