/**
=============================================================================
Smartport Virtual Hard Drive, version 0.99.0 (beta)
smartport (aka Cbus) protocol interpreter low level routines

If you would like to be involved in this project, or simply get my 
documentation for your own smartport project, feel free to contact me at
pcedric@hotmail.com

Copyright 2010 Cédric PELTIER, pcedric@hotmail.com

This file is part of the Smartport VHD project

Smartport VHD is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 3

Smartport VHD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Smartport VHD. If not, see <http://www.gnu.org/licenses/>.
=============================================================================
*/

#include "spvhd.h"
#include "protocol.h"
#include "cache.h"
#include <string.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#ifdef CFG_DEBUG_USB
#include "usb_serial.h"
#endif
#ifdef CFG_DEBUG_SERIAL
#include <LUFA/Drivers/Peripheral/SerialStream.h>
#endif

/**
React to smartport's state changes
*/
void sp_main_state()
{
	static uint8_t s;
    #ifdef CFG_DEBUG_PUSH
    //sp_debug_lines();
    //SP_DEBUG_PUSH((SP_IF_OACK?'A':'a'));
    //SP_DEBUG_PUSH((SP_IF_IREQ?'R':'r'));
    #endif
    enum sp_status status = sp_status_read();
    if(g_sp_state.status!=status)
    {
        switch(status)
        {
            case SP_ENABLED:
                sp_state_enabled();
                break;
            case SP_REQUEST:
                sp_state_request();
                break;
            case SP_DISABLED:
                sp_state_disabled();
                break;
            case SP_RESET:
                sp_state_reset();
                break;
        }
    }
    sp_update_olstrb();    
    sp_update_odr2();
    sp_update_oint();
    //SP_ODR2_PIN = SP_ODR2;
    //PINB = (1<<PB0);
}

/**
Read the status of the smartbus
@return
 SP_RESET if the smartbus is reseted
 SP_DISABLED if the smartbus is not currently enabled
 SP_ENABLED if the smartbus is enabled and the host is not ready
 SP_REQUEST id the smartbus is enabled and the host is ready
*/
enum sp_status sp_status_read()
{
	 // Disabling the device if no image mounted
	 //if(g_sp_state.drive_max==0)
	 //	return SP_DISABLED;
	
	 // Disabling the device if an Apple 3.5 is daisy chained ! Wrong !
    //if( SP_IF_I35SEL )
    //	return SP_DISABLED;
    
	 if( SP_IF_RESET )
		  return SP_RESET;
	 if( SP_IF_REQUEST )
		  return SP_REQUEST;   
	 if( SP_IF_ENABLED )
		  return SP_ENABLED;
    return SP_DISABLED;
}

/**
called when we detect that smarport is enabled
*/
void sp_state_enabled()
{
    SP_DEBUG_PUSH('E');

    // disable potential chained 5.25 floppy
    SP_DO_ODR2ON;
    SP_DO_OINTON;
    //SP_DEBUG_PUSH('W');
    
    switch(g_sp_state.status)
    {
        case SP_DISABLED:
        case SP_RESET:
            SP_DEBUG_PUSH('e');
            // smartport is going to send data
            // keep waiting the smartport to be synchronized
            if(sp_handshake_opening()==0)
                sp_state_request();
            else
                g_sp_state.status = SP_ENABLED;
            break;
        case SP_ENABLED:
        case SP_REQUEST:
            g_sp_state.status = SP_ENABLED;
            #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
            if(gb_debug&CFG_DEBUG_MASK_BUS)
                printf_P(PSTR("SPB: SP_ENABLED\r\n"));
            #endif
            break;
    }
    
}

/**
called when we detect that smarport is requesting
*/
void sp_state_request()
{
    SP_DEBUG_PUSH('R');

    // disable potential chained 5.25 floppy
    SP_DO_ODR2ON;
    SP_DO_OINTON;
    //SP_DEBUG_PUSH('W');

    g_sp_state.status = SP_REQUEST;
    switch(g_sp_state.activity)
    {
        case SP_RECEIVE:
            sp_receive_start();
            break;
            
        case SP_SEND:
            sp_send_start();
            break;
    }
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug&CFG_DEBUG_MASK_BUS)
        printf_P(PSTR("SPB: SP_REQUEST\r\n"));
    #endif
}

/**
called when we detect that smarport is disabled
*/
void sp_state_disabled()
{
    SP_DEBUG_PUSH('D');
    // let other devices control RD and ACK lines, so disable outputs
    SP_DO_ORDDMUTE;
    SP_DO_OACKOFF;
    sp_update_olstrb();
    sp_update_odr2();
    sp_update_oint();
    
    // smartport entered disabled state
    g_sp_state.status = SP_DISABLED;
    // be ready when it will become enabled
    sp_receive_init();
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug&CFG_DEBUG_MASK_BUS)
        printf_P(PSTR("SPB: SP_DISABLED\r\n"));
    #endif
}

/**
called when we detect that smarport is reseting
*/
void sp_state_reset()
{
    SP_DEBUG_PUSH('*');
    // smartport entered reset state
    // reset ourself

    // disable potential chained 5.25 floppy
    SP_DO_ODR2ON;
    SP_DO_OINTON;
    //SP_DEBUG_PUSH('W');

    // let other devices control RD and ACK lines, so disable outputs
    SP_DO_ORDDMUTE;
    SP_DO_OACKOFF;

    sp_init();
    g_sp_state.status = SP_RESET;
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug&CFG_DEBUG_MASK_BUS)
        printf_P(PSTR("SPB: SP_RESET\r\n"));
    #endif
}

/**
update OLSTRB from ILSTRB
- when all virtuals drives are initialized to allow initialisation of following intelligent resident 
- when smartport is disabled to allow access to 5.25 daisy chained drives
*/
void sp_update_olstrb()
{
    //static uint8_t s_lstrb;
    if(g_sp_state.initialized==1  || !SP_IF_ENABLED)
    {
		if(SP_IF_ILSTRB)
		{
			SP_DO_OLSTRBON;
/*			if(s_lstrb!='V')
			{
				s_lstrb='V';
				SP_DEBUG_PUSH(s_lstrb);
			}	
*/		}
		else
		{
			SP_DO_OLSTRBOFF;
/*			if(s_lstrb!='v')
			{
				s_lstrb='v';
				SP_DEBUG_PUSH(s_lstrb);
			}	
*/		}
    }
}

/**
update ODR2 from IDR2 when smartport is disabled
*/
void sp_update_odr2()
{
    //static uint8_t s_dr2;
    if(!SP_IF_ENABLED)
    {
    	 if(SP_IF_IDR2)
    	 {
    	 	SP_DO_ODR2ON;
/*			if(s_dr2!='W')
			{
				s_dr2='W';
				SP_DEBUG_PUSH(s_dr2);
			}	
*/    	 }
    	 else
    	 {
    	 	SP_DO_ODR2OFF;
/*			if(s_dr2!='w')
			{
				s_dr2='w';
				SP_DEBUG_PUSH(s_dr2);
			}	
*/    	 }
    }
}

/**
update OINT from IINT when smartport is disabled
*/
void sp_update_oint()
{
	#if HARDWARE_V1 == 0
	//static uint8_t s_dr2;
	if(!SP_IF_ENABLED)
	{
    	 if(SP_IF_IINT)
    	 {
    	 	SP_DO_OINTON;
/*			if(s_dr2!='W')
			{
				s_dr2='W';
				SP_DEBUG_PUSH(s_dr2);
			}	
*/    	 }
    	 else
    	 {
    	 	SP_DO_OINTOFF;
/*			if(s_dr2!='w')
			{
				s_dr2='w';
				SP_DEBUG_PUSH(s_dr2);
			}	
*/    	 }
    }
    #endif
}

/**
Do a closing handshake
@return 0 if the handshake was done, 1 if it was not done
*/
uint8_t sp_handshake_closing()
{
    SP_DO_OACKOFF;
    for(;;)
    {
        if(!SP_IF_REQUEST)
        {
            SP_DEBUG_PUSH('C');
            return 0;
        }
        if(!SP_IF_ENABLED)
        {
            SP_DEBUG_PUSH('c');
            return 1;
        }
        SP_DEBUG_PUSH(',');
        _delay_us(1);
    }
    SP_DEBUG_PUSH('c');
    return 1;
}

/**
Do an opening handshake
@return 0 if the handshake was done, 1 if it was not done
*/
uint8_t sp_handshake_opening()
{
    SP_DO_OACKON;
    for(;;)
    {
        if(SP_IF_REQUEST)
        {
            SP_DEBUG_PUSH('O');
            return 0;
        }
        if(!SP_IF_ENABLED)
        {
            SP_DEBUG_PUSH('o');
            return 1;
        }
        SP_DEBUG_PUSH(',');
        //sp_debug_lines();
        _delay_us(1);//1
    }
    SP_DEBUG_PUSH('o');
    return 1;
}

