/**
=============================================================================
Smartport Virtual Hard Drive, version 0.99.0 (beta)
smartport (aka Cbus) protocol interpreter

If you would like to be involved in this project, or simply get my 
documentation for your own smartport project, feel free to contact me at
pcedric@hotmail.com

Copyright 2010 Cédric PELTIER, pcedric@hotmail.com

This file is part of the Smartport VHD project

Smartport VHD is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 3

Smartport VHD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Smartport VHD. If not, see <http://www.gnu.org/licenses/>.
=============================================================================
*/


#include "spvhd.h"
#include "protocol.h"
#include "cache.h"
#include "ds1307.h"
#include <string.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "extsram.h"
#include "usb_storage.h"
#ifdef CFG_DEBUG_USB
#include "usb_serial.h"
#endif
#include <Lib/FATFs/ff.h>



/**
Be ready to receive data from the host
*/
void sp_receive_init()
{
    SP_DEBUG_PUSH('l');
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    // Too long for smartport timing
    //if(gb_debug & CFG_DEBUG_MASK_RCV) printf_P(PSTR("SPB: sp_receive_init()\r\n"));
    #endif

    // be ready to listen the host
    
    cli();
    
    // RD line at high impedance until we need to talk to the host
    SP_DO_ORDDMUTE;

    // timer 1 interrupts disabled for now
    TIMSK1 = 0;
    // clearing pending interrupt
    TIFR1=(1<<OCIE1C)|(1<<ICF1);//(1<<TOIE1);
    // don't start the timer now
    TCCR0A = 0;
    TCCR1A = 0; 
    TCCR1B = 0; 
    OCR1C=0;
    ICR1=0;
    // Set the counter 1 start value
    TCNT1 = 0;

	// use the timer 1 with a /64 prescaler to sense the WR line on a 4us basis
	TCCR1B = ((1<<CS11)|(1<<CS10));

    sei();          

    sp_init_packet(0,0,0,g_sp_buffer,SP_BUFRAWSIZE);

    // start the reception now, don't wait the request signal
    // Enable interrupt for event capture on timer 1 
    // in order to start reception on the first msb of the packet
    //    sp_receive_start()
    
    // if smartport enabled, start the opening handshake now !
    if(SP_IF_ENABLED)
    	 SP_DO_OACKON;
    
    SP_DEBUG_PUSH('L');
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug & CFG_DEBUG_MASK_RCV)        printf_P(PSTR("SPB: sp_receive_start()\r\n"));
    #endif
}

/**
Enable reception of data from the host
*/
void sp_receive_start()
{
	TIMSK1 = (1<<ICIE1);
}

/**
Manage the reception of data from the smartport
*/
void sp_receive_main()
{
    struct sp_states* p_state = &g_sp_state;
	if(p_state->receive_step == SP_END)
	{
		// we have an incoming packet
		if(p_state->error == SP_ERROK)
			sp_receive_packet();
		
		if(p_state->receive_step == SP_GOT)
		{
			// we received a write request, must receive the data packet next time
			// the pointer to the buffer elected to receive the data must be keept
			uint8_t* p1 = p_state->p_buffer;
			uint8_t* p2 = p_state->p_buffer_end;
			sp_receive_init();
			p_state->p_buffer = p1;
			p_state->p_buffer_end = p2;
			SP_DO_OACKON;	
		}
		else if(p_state->error != SP_ERROK)
		{
			// an error occured, restart the reception
			sp_receive_init();
			SP_DO_OACKON;
		}   
	}
}

/**
Got a packet to manage
*/
void sp_receive_packet()
{
    struct sp_states* p_state = &g_sp_state;
    CSET_BANK_SET;
    uint8_t command = p_state->p_buffer[0];
    CSET_BANK_UNSET;
    if( (p_state->drive_current!=255 || ( p_state->initialized == 0 && command==SP_CMDSID)))// && g_sp_state.error==SP_ERROK ) //p_state->drive_set<p_state->drive_max
    {
        // this packet is for us
        // finish the closing handshake before the 120us of the host timeout
        // the host must receive the packet's begin mark of the response whithin the next 960us
        // note: the handshake is alreay initiated by the isr
        // note: don't waste our precious time, prepare the response and then finish the closing handshake [wrong]
        if(sp_handshake_closing()==0)
        {
			  //die(LEDMASK_POWER,LEDMASK_WARNING,LEDMASK_ERROR,250);
			  switch(g_sp_packet.header.pkttype)
			  {
					case SP_PKTCMD:
						 sp_receive_command(p_state);
						 break;
					/* used only by devices ... ?
					case SP_PKTSTA:
						 sp_receive_status();
						 break;
					*/
					case SP_PKTDAT:
						 sp_receive_data(p_state);
						 break;
					 default:
						#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
						if(gb_debug & CFG_DEBUG_MASK_RCV)							printf_P(PSTR("SPB: got unknown packet type %02X\r\n"),g_sp_packet.header.pkttype);
						#endif
						sp_receive_init();
			  }
		  }
		  /*
		  else
		  {
		  	  sp_receive_init();
		  }
		  */
    }
    else
    {
        // this packet is definitively not for us
        // ...or it is malformed...
        // don't do the handshake, the host will retry if required
        SP_DEBUG_PUSH('%');
        #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
        if(gb_debug & CFG_DEBUG_MASK_RCV)            printf_P(PSTR("SPB: got a packet not for us\r\n"));
        #endif
        sp_receive_init();
    }
}

/**
Manage the reception of a command packet
*/
void sp_receive_command(struct sp_states* p_state)
{
    // set to 1 to send a response, 0 to receive another packet
    uint8_t b_send=0;
    
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug & CFG_DEBUG_MASK_RCV)        printf_P(PSTR("SPB: Got a command packet\r\n"));
	#endif
	CSET_BANK_SET;
	uint8_t command = p_state->p_buffer[0];
	CSET_BANK_UNSET;
    switch(command)
    {
        case SP_CMDSID:
			b_send = sp_receive_id();
            break;
        case SP_CMDSTATUS:
            b_send = sp_receive_status(p_state);
            break;
        case SP_CMDREAD:
            b_send = sp_receive_read();
            break;
        case SP_CMDWRITE:
            b_send = sp_receive_write();
            break;
        case SP_CMDFORMAT:
            // required only for amovible media
            b_send = 1;
            sp_send_status(SP_ERR_OK);
            break;
        case SP_CMDCTRL:
            b_send = sp_receive_control(p_state);
            break;
			#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
		 default:
			if(gb_debug & CFG_DEBUG_MASK_RCV)				printf_P(PSTR("SPB: got unknown command type %02X\r\n"),command);
			#endif
    }
    
	  //if(!SP_IF_REQUEST)
	  if(sp_handshake_closing()==0)
	  {
			SP_DEBUG_PUSH('H');
			if(b_send==1)
			{
				// a status packet is to be send
				sp_send_main();
			}
			else
			{
				// fall back into reception mode
			}
	  }
	  else
	  {
			SP_DEBUG_PUSH('h');
	  }
}

/**
Manage an id set
*/
uint8_t sp_receive_id()
{
    SP_DEBUG_PUSH('I');
    struct sp_states* p_state = &g_sp_state;
    // got a smartport id, thanks !

    //if(g_sp_state.initialized == 0)
    //if(USB_HostState != HOST_STATE_Configured)
    if(p_state->drive_max == 0)
    {
    	 // We can safely initialize usb layer and enumerate virtual drive here
    	 // as the computer is waiting indefinitively our response !
    	 sp_init_drives();
	 }
	
    if(p_state->drive_set < p_state->drive_max)
    {
    	// store the id for the current virtual drive
		p_state->drive_current = p_state->drive_set;
		++p_state->drive_set;
		p_state->id[p_state->drive_current] = g_sp_packet.header.iddst;
		#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
		if(gb_debug & CFG_DEBUG_MASK_RCV) printf_P(PSTR("SPB: Got id %d for the drive %d, drive max %d\r\n"),p_state->id[p_state->drive_current],p_state->drive_current,p_state->drive_max);
		/*
		printf_P(PSTR("id packet %02X %02X %02X %02X %02X\r\n"),g_sp_state.p_buffer[0],g_sp_state.p_buffer[1],g_sp_state.p_buffer[2],g_sp_state.p_buffer[3],g_sp_state.p_buffer[4]);
		for(uint8_t k=0;k<state->drive_set;k++)
		{
			printf_P(PSTR("Index %u ID %u DIB %02X %02X %02X %02X %02X ... %02X %02X\r\n"),k,state->id[k],state->dib[k][0],state->dib[k][1],state->dib[k][2],state->dib[k][3],state->dib[k][4],state->dib[k][21],state->dib[k][22],state->dib[k][23],state->dib[k][24]);
		}
		*/
		#endif
		// prepare the response, 
		// if all virtual drive got there id, allow initialization of chained residents
		uint8_t status = SP_ERR_OK;
		if(p_state->drive_set >= p_state->drive_max)
		{
			p_state->initialized=1;
			
			// if all virtual drives are initialized,
			// update the status accordingly to the sense on OLSTRB
			// 0x00 if low (another resident is chained)
			// 0x21 if high (disk II or nothing more in the chain)
			if(SP_IF_IHDSEL)
				status = SP_ERR_IOERROR;		
		}
		sp_send_status(status);
		return 1;
	}
	return 0;
}

/**
Manage a read request
*/
uint8_t sp_receive_read()
{
	SP_DEBUG_PUSH('<');
	
	// make local copy to speed up code and reduce size
	struct sp_states* p_state = &g_sp_state;
	
	// Select the right file image
	//sp_select_image(p_state);
	
	// attempt to read the requested data from the medium
	CSET_BANK_SET;
	uint32_t address = sp_size_get( (struct sp_size*)(&p_state->p_buffer[4]) );
	CSET_BANK_UNSET;
	// current block to read
	p_state->block_address = address;
	#ifdef CFG_READ_AHEAD
	// next block to read for the read ahead cache system
	p_state->block_address_next = address+1;
	#endif
	
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_RCV) printf_P(PSTR("SPB: drive %u, read block %lu, position %lu\r\n"),p_state->drive_current,p_state->block_address,(DWORD)p_state->block_address*(DWORD)SP_BLOCKSIZE);
	#endif
    
	// find the block in the cache memory
	uint8_t* pointer = cset_get(p_state->drive_current,p_state->block_address,1);
	if(pointer == 0)
	{
    	// not found in the cache, need to retreive a buffer and load the block
    	pointer = cset_new(p_state->drive_current,p_state->block_address,1);
    	FIL* drive = &p_state->drive[p_state->drive_current];
    	//FIL* drive = &p_state->drive_descriptor;
		#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
		if(gb_debug & CFG_DEBUG_MASK_RCV) printf_P(PSTR("SPB: drive pointer %04X\r\n"),drive);
	   #endif

		FRESULT rc = f_lseek(drive,(DWORD)(p_state->block_address)*(DWORD)(SP_BLOCKSIZE));
		if(rc==FR_OK)
		{
			UINT size_read=0;
			CSET_BANK_SET;
			rc = f_read(drive,pointer,SP_BLOCKSIZE,&size_read);
			CSET_BANK_UNSET;
			#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
			//printf_P(PSTR("SPB: read size %lu %d\r\n"),size_read);
			#endif
			
		}
		if(rc!=FR_OK)
		{
			#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
			printf_P(PSTR("SPB: reading error %d\r\n"),rc);
			#endif
			sp_send_status(SP_ERR_IOERROR);
			return 1;
		}
	}
	
    // prepare the response
    sp_send_data(pointer);

	if(g_sp_packet.canary!=SP_CANARY)
    {
        #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
        printf_P(PSTR("SPB: packet buffer scrambled %04X\r\n"),g_sp_packet.canary);
        #endif
        die(LEDMASK_SP,LEDMASK_SP|LEDMASK_ERROR,LEDMASK_SP|LEDMASK_ERROR,1000);
    }
    
    return 1;
}

/**
Manage a write request
*/
uint8_t sp_receive_write()
{
	SP_DEBUG_PUSH('>');
	
	// make local copy to speed up code and reduce size
	struct sp_states* p_state = &g_sp_state;
	
	// store the block address waiting data
	CSET_BANK_SET;
	p_state->block_address = sp_size_get( (struct sp_size*)(&p_state->p_buffer[4]) );
	CSET_BANK_UNSET;
	
	// find the block in the cache memory
	uint8_t* pointer = cset_get(p_state->drive_current,p_state->block_address,1);
	if(pointer == 0)
	{
		pointer = cset_new(p_state->drive_current,p_state->block_address,1);
	}
	
	p_state->p_buffer = pointer;
	p_state->p_buffer_end = pointer+SP_BLOCKSIZE;
	
	// start a new reception after this packet
	p_state->receive_step = SP_GOT;

	// remember we are receiving a data block
	p_state->ctrlcode = SPVHD_BLOCK;
	
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_RCV)	  printf_P(PSTR("SPB: write block %lu\r\n"),p_state->block_address);
	#endif
	
	return 0;
}

/**
Manage the reception of a data packet
*/
void sp_receive_data(struct sp_states* p_state)
{
	SP_DEBUG_PUSH('\\');
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_RCV)	  printf_P(PSTR("SPB: receive data %02X\r\n"),p_state->ctrlcode);
	#endif

	// Select the right file image
	//sp_select_image(p_state);
//	sp_receive_data_block(p_state);
	if(p_state->ctrlcode > SPVHD_BLOCK)
	{
	}
	else
	{
		switch(p_state->ctrlcode)
		{
			case SPVHD_BLOCK:
				sp_receive_data_block(p_state);
				break;
			case SPVHD_DATE:
				// set date and hour
				sp_receive_data_date(p_state);
				break;
			case SPVHD_EXPLORE:
				// set pathname to explore
				sp_receive_data_explore(p_state);
				break;
			case SPVHD_PARMSEL:
				// select the current vitual device
				sp_receive_data_vdsel(p_state);
				break;
			case SP_CTRLDCB:
				// store the dcb
				p_state->devnum = 0;
			case SPVHD_PARMBOOT:
				// store configuration for the curent virtual device
				sp_receive_data_parm(p_state);
				break;
		}
	}	
	if(sp_handshake_closing()==0)
	{
	  SP_DEBUG_PUSH('H');
	  sp_send_main();
	}
	else
	{
	  SP_DEBUG_PUSH('h');
	  sp_receive_init();
	}

}

/**
Manage the reception of a data block to write to media
*/
void sp_receive_data_block(struct sp_states* p_state)
{
	uint8_t l_rc2 = SP_ERR_IOERROR;
	
	// Write a data block
	FIL* lp_drive = &p_state->drive[p_state->drive_current];
	
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
		if(gb_debug & CFG_DEBUG_MASK_RCV) printf_P(PSTR("SPB: Got a data packet to write at %lu dib=%02X\r\n"),p_state->block_address,p_state->dib[p_state->drive_current][0]);
	#endif
	
	if( (p_state->dib[p_state->drive_current][0]&4) == 0)
	{
		// Write allowed
		FRESULT l_rc = f_lseek(lp_drive,(DWORD)(p_state->block_address)*(DWORD)(SP_BLOCKSIZE));    
		if(l_rc==FR_OK)
		{
			UINT l_size_write=0;
			CSET_BANK_SET;
			l_rc = f_write(lp_drive,p_state->p_buffer,SP_BLOCKSIZE,&l_size_write);
			CSET_BANK_UNSET;
		}
		
		if(l_rc==FR_OK)
		{
		  l_rc2 = SP_ERR_OK;
		}
		else
		{
		  #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
		  printf_P(PSTR("SPB: writing error %d\r\n"),l_rc);
		  #endif
		  // Maybe the disk image file is read only...
		  if(l_rc == FR_DENIED)
		  	  l_rc2 = SP_ERR_NOWRITE;
		}
	}
	else
	{
		// Disk image is declared read only
		l_rc2 = SP_ERR_NOWRITE;
	}
	sp_send_status(l_rc2);
}

/**
Manage the reception of a date to set the rtc
*/
void sp_receive_data_date(struct sp_states* p_state)
{
	uint8_t l_data[9];
	uint8_t* lp_ptr = p_state->p_buffer; 
   bool l_rc;
   
	CSET_BANK_SET;
	
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_RCV) printf_P(PSTR("SPB: receive date %02X %02X %02X %02X\r\n"),lp_ptr[0],lp_ptr[1],lp_ptr[2],lp_ptr[3]);
	#endif

   CSET_BANK_UNSET;
   
	l_rc = ds1307_read(l_data);
	if(l_rc)
	{
		l_data[1] = bin2bcd(lp_ptr[2]);
		l_data[2] = (l_data[2] & 0xC0) | (bin2bcd(lp_ptr[3]) & 0x3F);
		l_data[4] = bin2bcd(lp_ptr[0] & 31);	
		l_data[5] = bin2bcd( (lp_ptr[0] >> 5) | ((lp_ptr[1]&1)<<3 ) );
		l_data[6] = bin2bcd(lp_ptr[1] >> 1);
		l_rc = ds1307_write(l_data);
	}
	sp_send_status(l_rc?SP_ERR_OK:SP_ERR_IOERROR);
}

/**
Manage the reception of a folder to explore within the repository
*/
void sp_receive_data_explore(struct sp_states* p_state)
{
	FRESULT l_rc;
	uint8_t i;
	uint8_t* lp_ptr;
	uint8_t l_size;
	struct spvhd_explore* lp_spvhd_explore;
	// Reuse the end of the packet data buffer for our own use
   TCHAR* lp_path = (TCHAR*)&g_sp_packet.data[SP_BUFDATSIZE-128-11]
	CSET_BANK_SET;
	
	lp_spvhd_explore = (struct spvhd_explore*)p_state->p_buffer;
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	//if(gb_debug & CFG_DEBUG_MASK_RCV) printf_P(PSTR("SPB: receive explore %02X %c %c %c %c\r\n"),lp_spvhd_explore->pathsize,lp_spvhd_explore->path[0],lp_spvhd_explore->path[1],lp_spvhd_explore->path[2],lp_spvhd_explore->path[3]);
	#endif
	i = 0;
	lp_ptr = lp_spvhd_explore->path;
	l_size = lp_spvhd_explore->pathsize; 
	// add a path separator at the start if required
	if(l_size == 1)
	{
		if(lp_ptr[0]=='/')
			l_size = 0;
	}
	else if(l_size > 1)
	{
		if(lp_ptr[0]!='/')
		{
			// insert a path separator at the first position
			for(i=l_size+1;i>0;i--)
			{
				lp_ptr[i] = lp_ptr[i-1]; 
			}
			lp_ptr[0] = '/';
			l_size++;
		}
		// remove the ending path separator
		if(lp_ptr[l_size-1] == '/')
			l_size--;
	}
	// make a zero string
	lp_ptr[l_size]=0;
	sprintf_P(lp_path,PSTR("0:/spvhd%s"),lp_ptr);
	if(lp_spvhd_explore->pathsize == 0)
		lp_path[8]=0;
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_RCV) printf_P(PSTR("SPB: receive explore %d %s %s\r\n"),lp_spvhd_explore->pathsize,lp_spvhd_explore->path,lp_path);
	#endif
	
	// get directory infos
	l_rc = f_opendir(&p_state->directory,lp_path);

	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_RCV) printf_P(PSTR("SPB: f_opendir() %d\r\n"),l_rc);
	#endif
   
	CSET_BANK_UNSET;

  	sp_send_status(l_rc==FR_OK?SP_ERR_OK:SP_ERR_IOERROR);
}

/**
Manage the reception of a virtual device selection
*/
void sp_receive_data_vdsel(struct sp_states* p_state)
{
	uint8_t l_rc = SP_ERR_IOERROR
	CSET_BANK_SET;
	if(p_state->p_buffer[0]<=SP_MAX_DRIVE)
	{
		p_state->devnum = p_state->p_buffer[0];
		l_rc = SP_ERR_OK;
	}
   CSET_BANK_UNSET;
	
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_RCV) printf_P(PSTR("SPB: select device %u\r\n"),p_state->devnum);
	#endif

	sp_send_status(l_rc);
}

/**
Manage the reception of a parameter block for a virtual device
*/
void sp_receive_data_parm(struct sp_states* p_state)
{
	uint8_t l_rc;
	uint8_t lp_pathname[22];
	uint8_t* lp_src;
	uint8_t* lp_dst;
   
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_RCV)
	{
		if(p_state->devnum>0)
		{
			struct spvhd_parms* lp_spvhd_parm = (struct spvhd_parms*)p_state->p_buffer; 
			printf_P(PSTR("SPB: receive parm %02X %02X %02X %02X\r\n"),lp_spvhd_parm->parms,lp_spvhd_parm->pathsize,lp_spvhd_parm->path[0],lp_spvhd_parm->path[1]);
		}
		else
		{
			puts_P(PSTR("SPB: receive dcb"));
		}
	}
	#endif
  
	// determine the path to the configuration file
	strcpy_P(lp_pathname,PSTR("0:/spvhd/"));
	lp_dst = &lp_pathname[strlen(lp_pathname)];
	lp_src = PSTR("cfgboot.dat");
	strcpy_P(lp_dst,lp_src);
	
	l_rc = SP_ERR_IOERROR;
	if( cfg_file_io(lp_pathname, p_state->devnum, 1, p_state->p_buffer) == 0)
	  	l_rc = SP_ERR_OK;
	
	sp_send_status(l_rc);
}

/**
Manage the reception of a status packet
*/
uint8_t sp_receive_status(struct sp_states* p_state)
{
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug & CFG_DEBUG_MASK_RCV) printf_P(PSTR("SPB: Got status cmd\r\n"));
    #endif
    CSET_BANK_SET;
    uint8_t l_statcode = p_state->p_buffer[4];
    CSET_BANK_UNSET;
    switch(l_statcode)
    {
        case SP_STASTA:
            // device status request
            SP_DEBUG_PUSH('u');
            sp_send_ds(SP_ERR_OK);
            return 1;
        case SP_STADCB:
            // device control block request
            SP_DEBUG_PUSH('v');
            // send the dcb only for the lowest smartport id we are managing
            // this is easier to detect chained smartportvhd
            if(sp_lowest_id_get(p_state) == g_sp_packet.header.iddst)
            	sp_send_dcb(p_state);
            else
            	sp_send_status(SP_ERR_IOERROR);
            return 1;
        case SP_STADIB:
            // device information block request
            SP_DEBUG_PUSH('w');
            sp_send_dib(SP_ERR_OK);
            return 1;
        case SPVHD_DATE:
        	  	// date request
            SP_DEBUG_PUSH('x');
           	sp_send_date(p_state);
            return 1;
        case SPVHD_NEXTFILE:
        	  	// next file name for parsing repository
            SP_DEBUG_PUSH('x');
           	sp_send_nextfile(p_state);
            return 1;
        case SPVHD_PARMBOOT:
				// get virtual device parameter 	  
				SP_DEBUG_PUSH('x');
				sp_send_parm(p_state,l_statcode);
				return 1;
        default:
            SP_DEBUG_PUSH('y');
            sp_send_status(SP_ERR_BADCMD);
            #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
            if(gb_debug & CFG_DEBUG_MASK_RCV) printf_P(PSTR("SPB: %02X %02X %02X %02X <%02X> %02X\r\n"),g_sp_state.p_buffer[0],g_sp_state.p_buffer[1],g_sp_state.p_buffer[2],g_sp_state.p_buffer[3],g_sp_state.p_buffer[4],g_sp_state.p_buffer[5]);
            #endif
            return 1;
    }
    return 0;
}

/**
Manage the reception of a control packet
*/
uint8_t sp_receive_control(struct sp_states* p_state)
{
    CSET_BANK_SET;
    uint8_t l_ctrlcode = p_state->p_buffer[4];
    CSET_BANK_UNSET;
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug & CFG_DEBUG_MASK_RCV)        printf_P(PSTR("SPB: Got control cmd %02X\r\n"),l_ctrlcode);
    #endif
    switch(l_ctrlcode)
    {
        case SP_CTRLRST:
            // soft reset request
            SP_DEBUG_PUSH('v');
            sp_send_status(SP_ERR_BADCMD);
            return 1;
        case SP_CTRLNLS:
            // set new line status (char device only)
            SP_DEBUG_PUSH('w');
            sp_send_status(SP_ERR_BADCMD);
            return 1;
        case SP_CTRLINT:
            // service device interrupt
            SP_DEBUG_PUSH('x');
            sp_send_status(SP_ERR_BADCMD);
            return 1;
         case SP_CTRLDCB:
         	// set dcb
        case SPVHD_DATE:
        	  	// set date request
        case SPVHD_EXPLORE:
        	  // start repository parsing request
        case SPVHD_PARMSEL:
        	  // get the virtual device number to work on
        case SPVHD_PARMBOOT:
				// get virtual device parameter 	  
            SP_DEBUG_PUSH('y');
        	  return sp_receive_dcb(p_state,l_ctrlcode);
    }
	SP_DEBUG_PUSH('!');
	sp_send_status(SP_ERR_BADCMD);
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_RCV)		 printf_P(PSTR("SPB: %02X %02X %02X %02X <%02X> %02X\r\n"),g_sp_state.p_buffer[0],g_sp_state.p_buffer[1],g_sp_state.p_buffer[2],g_sp_state.p_buffer[3],g_sp_state.p_buffer[4],g_sp_state.p_buffer[5]);
	#endif
	return 1;
}

/**
Manage a control dcb request
*/
uint8_t sp_receive_dcb(struct sp_states* p_state,uint8_t l_ctrlcode)
{
	SP_DEBUG_PUSH('>');
	
	/// @todo: use the end of the packet buffer instead of reserve a full block
	// requisite a buffer to receive the dcb
	uint8_t* lp_pointer = cset_new(0xFF,0xFFFFFFFF,0);
	
	p_state->p_buffer = lp_pointer;
	p_state->p_buffer_end = lp_pointer+SP_BLOCKSIZE;
	
	// start a new reception after this packet
	p_state->receive_step = SP_GOT;
	
	// remember we are receiving a dcb
	p_state->ctrlcode = l_ctrlcode;
	
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_RCV)	  printf_P(PSTR("SPB: receive dcb %u\r\n"),l_ctrlcode);
//struct sp_packet* p_packet = &g_sp_packet;
//printf_P(PSTR("SPB: <%02X %02X><%02X><%02X %02X><%02X %02X %02X %02X %02X %02X %02X %02X %02X>\r\n"),p_packet->header.pkttype,p_packet->header.pktauxtype,p_packet->header.status,p_packet->header.nbodd,p_packet->header.nbgrp,p_state->p_buffer[0],p_state->p_buffer[1],p_state->p_buffer[2],p_state->p_buffer[3],p_state->p_buffer[4],p_state->p_buffer[5],p_state->p_buffer[6],p_state->p_buffer[7],p_state->p_buffer[8]);		
	#endif
	
	return 0;
}

/**
Open the right image file related to the requested virtual drive
*/
/*
void sp_select_image(struct sp_states* p_state)
{
	if(p_state->drive_current != p_state->drive_descriptor_index)
	{
		// The requested drive isn't related to the current opened image file
		if(p_state->drive_descriptor_index != 255)
		{
			// Closing the opened drive if required
			#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
			if(gb_debug & CFG_DEBUG_MASK_RCV)				 printf_P(PSTR("SPB: Closing image for drive %u\r\n"),p_state->drive_descriptor_index);
			#endif
			f_close(&p_state->drive_descriptor);
			p_state->drive_descriptor_index = 255;
		}
		// Opening the requested drive
		uint8_t filename[24];
		sprintf(filename,"0:/spvhd/drive%d.img",p_state->drive_current);
		#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
		if(gb_debug & CFG_DEBUG_MASK_RCV)			 printf_P(PSTR("SPB: Opening image %s for drive %u\r\n"),filename,p_state->drive_current);
		#endif
		uint8_t rc = f_open(&p_state->drive_descriptor,filename,FA_OPEN_EXISTING | FA_READ | FA_WRITE);
		if(rc == FR_OK)
		{
			p_state->drive_descriptor_index = p_state->drive_current;
		}
		else
		{
			#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
			if(gb_debug & CFG_DEBUG_MASK_RCV)				printf_P(PSTR("SPB: Error opening image\r\n"));
			#endif
		}		
	}
}
*/
