/**
=============================================================================
Smartport Virtual Hard Drive, version 1.00.00
smartport (aka Cbus) protocol interpreter

If you would like to be involved in this project, or simply get my 
documentation for your own smartport project, feel free to contact me at
pcedric@hotmail.com

Copyright 2010,2011,2012 Cédric PELTIER, pcedric@hotmail.com

This file is part of the Smartport VHD project

Smartport VHD is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 3

Smartport VHD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Smartport VHD. If not, see <http://www.gnu.org/licenses/>.
=============================================================================
*/


#include "spvhd.h"
#include "protocol.h"
#include "cache.h"
#include <string.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "extsram.h"
#include "usb_storage.h"
#ifdef CFG_DEBUG_USB
#include "usb_serial.h"
#endif

/**
Called at the first incoming bit from the host
Receive a full packet, then let the application continue
This is required as the AVR is not power enough to manage
the smartport asynchronously
*/
ISR(TIMER1_CAPT_vect)
{
	// start of the critical timing phase

	// 44 cycles are lost by the interrupt call,
	// we have only 64 cycles between bits @16MHz,
	// get the current state of the line to detect incoming bits
	g_sp_state.receive_line = (SP_IWRD_PIN & SP_IWRD);

    SP_DEBUG_PUSH('+');
	// we are concerned only by smartport requests (this is managed by the upper layer)
	//if(SP_IF_REQUEST)
	{
		// interrupts are disabled globaly by the interrupt call, 
		// no need to disabled our interrupt now
		
		// update visual indicator of smartport activity
		LEDs_TurnOnLEDs(LEDMASK_SP);	

		// receive synchronisation sequence and wait the start of the packet
		while(sp_get_byte()!=SP_PKTBEGIN)
		{
			//SP_DEBUG_PUSH(';');
			// check for smartport presence
			if(!SP_IF_ENABLED)
			{
				sp_receive_isr_end();
				g_sp_state.error = SP_ERRRESET;
				SP_DEBUG_PUSH('9');
				return;
			}
		}
		
		// receive header
		uint8_t* p_to = g_sp_state.p_buffer;
		uint8_t* p_packet = g_sp_state.p_packet;
		uint8_t* p_to_end = g_sp_state.p_buffer_end;
		uint8_t crc_computed = 0;
		//SP_DEBUG_PUSH(':');
		for(uint8_t counter=7;counter>0;counter--)
		{
			uint8_t byte = sp_get_byte();
			// check for smartport presence
			if(!SP_IF_ENABLED)
			{
				sp_receive_isr_end();
				g_sp_state.error = SP_ERRRESET;
				SP_DEBUG_PUSH('9');
				return;
			}
			//SP_DEBUG_PUSH('9');
			// update the computed crc and the packet header
			crc_computed ^= byte;
			byte &= (uint8_t)127; 
			*p_packet++ = byte;
		}
		
		// select the bank of memory used for caching blocks
		CSET_BANK_SET;

		// receive odds bytes
		if(g_sp_packet.header.nbodd>0)
		{
			uint8_t msb = sp_get_byte();
			//*p_packet++ = msb;
			//SP_DEBUG_PUSH('1');
			for(uint8_t counter=g_sp_packet.header.nbodd;counter>0;counter--)
			{
				//SP_DEBUG_PUSH('2');
				uint8_t byte = sp_get_byte();
				// check for smartport presence
				if(!SP_IF_ENABLED)
				{
					sp_receive_isr_end();
					g_sp_state.error = SP_ERRRESET;
					SP_DEBUG_PUSH('9');
					return;
				}
				//*p_packet++ = byte;
				msb = (msb<<1);
				byte = (byte&(uint8_t)(127))|(msb&(uint8_t)(128));
				crc_computed ^= byte;
				*p_to++ = byte;
			}
		}
		
		// receive group of seven bytes encoded as height bytes
		for(uint8_t counter2=g_sp_packet.header.nbgrp;counter2>0;counter2--)
		{
			// some group to receive
			uint8_t msb = sp_get_byte();
			// check for smartport presence
			if(!SP_IF_ENABLED)
			{
				sp_receive_isr_end();
				g_sp_state.error = SP_ERRRESET;
				SP_DEBUG_PUSH('9');
				return;
			}
			//*p_packet++ = msb;
			//SP_DEBUG_PUSH('3');
			for(uint8_t counter=7;counter>0;counter--)
			{
				//SP_DEBUG_PUSH('4');
				uint8_t byte = sp_get_byte();
				//*p_packet++ = byte;
				msb = (msb<<1);
				byte = (byte&(uint8_t)(127))|(msb&(uint8_t)(128));
				crc_computed ^= byte;
				*p_to = byte;
				// in this loop, we can trash memory over the end of the buffer
				// so better to check it
				if(p_to<p_to_end)
					++p_to;
			}
		}
		
		// select the default bank
		// all data are now stored
		CSET_BANK_UNSET;

		// receive first crc byte
		uint8_t tmp_crc1 = sp_get_byte();
		// check for smartport presence
		if(!SP_IF_ENABLED)
		{
			sp_receive_isr_end();
			g_sp_state.error = SP_ERRRESET;
			SP_DEBUG_PUSH('9');
			return;
		}
		*p_packet++ = tmp_crc1;
		//g_sp_state.crc_computed = crc_computed;
		
		// receive last crc byte
		uint8_t tmp_crc2 = sp_get_byte();
		// check for smartport presence
		if(!SP_IF_ENABLED)
		{
			sp_receive_isr_end();
			g_sp_state.error = SP_ERRRESET;
			SP_DEBUG_PUSH('9');
			return;
		}
		*p_packet++ = tmp_crc2;
		tmp_crc2 = tmp_crc2 << 1;
		tmp_crc2 |= 1;
		uint8_t crc_received = tmp_crc1 & tmp_crc2;
		// receive end packet mark
		tmp_crc1 = sp_get_byte();
		*p_packet++ = tmp_crc1;
		//g_sp_state.crc_received = crc_received;
		// receive the dummy byte
		tmp_crc2 = sp_get_byte();
		*p_packet++ = tmp_crc2;
		
		if(crc_received != crc_computed)
		{
			g_sp_state.error = SP_ERRCRC;
			SP_DEBUG_PUSH('-');
		}
		else if(tmp_crc1 != SP_PKTEND)
		{
			g_sp_state.error = SP_ERRPACKET;
		}
		else
		{
			// start the closing handshake now to avoid possible synchronization problem
			// only if the packet is for us
			struct sp_states* p_state = &g_sp_state; 
			uint8_t iddst = g_sp_packet.header.iddst;
			uint8_t drive_set = p_state->drive_set;
			//uint8_t drive_max = p_state->drive_max;
			uint8_t drive_current = 255;
		
			// find the destination drive
			for(uint8_t i=0;i<drive_set && drive_current==255;i++)
				if(p_state->id[i]==iddst)
					drive_current = i;
	
			p_state->drive_current = drive_current;
			CSET_BANK_SET;
			uint8_t command = p_state->p_buffer[0];
			CSET_BANK_UNSET;
			if( (drive_current!=255 || ( p_state->initialized == 0 && command==SP_CMDSID)) && g_sp_state.error==SP_ERROK ) //drive_set<drive_max
			{
				SP_DO_OACKOFF;
			}
			if(drive_current == 255)
			{
				#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
				struct sp_packet* p_packet = &g_sp_packet;
				printf_P(PSTR("SPB: %02X->%02X <%02X %02X><%02X><%02X %02X><%02X %02X %02X %02X %02X %02X %02X %02X>\r\n"),p_packet->header.idsrc,iddst,p_packet->header.pkttype,p_packet->header.pktauxtype,p_packet->header.status,p_packet->header.nbodd,p_packet->header.nbgrp,command,p_state->p_buffer[1],p_state->p_buffer[2],p_state->p_buffer[3],p_state->p_buffer[4],p_state->p_buffer[5],p_state->p_buffer[6],p_state->p_buffer[7]);		
				#endif
			}
		}
		// end of the critical timing phase
		
		#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
		if(crc_received != crc_computed)
		{
			printf_P(PSTR("SPB: crc error %02X %02X\r\n"),crc_received, crc_computed);
			sp_debug_packet();
			#if defined(CFG_DEBUG_USB)
			usb_main();
			#endif
		}
		else if(tmp_crc1 != SP_PKTEND)
		{
			printf_P(PSTR("SPB: end packet error <%02X>\r\n"),tmp_crc1);
			sp_debug_packet();
			#if defined(CFG_DEBUG_USB)
			usb_main();
			#endif
		}
		#endif

		sp_receive_isr_end();
	}
}

/**
Read a byte from the smartport synchronously
Timer1 Input capture is used to synchronize reception on the msb
Timer1 /64 is then used to catch the following 7 bits on a 4us basis
*/
uint8_t sp_get_byte()
{
	// waiting msb to synchronize the reception...
	uint8_t line = g_sp_state.receive_line;
	// force the loop statement to be 8 bits computing
	uint8_t tmp;
	while( (tmp=(SP_IWRD_PIN & SP_IWRD)) == line )
	{
		// check for smartport presence
		if(!SP_IF_ENABLED)
			return 0;
	}
	// reset the prescaler so the first increment will be a full period
	GTCCR = (1<<PSRSYNC);
	// reset timer value
	TCNT1L = 0;
	// update the line state
	line ^= SP_IWRD;
	// set timer value to detect timer change
	uint8_t last = 0;
	// current bit to be received
	uint8_t bit = 0x40;
	// current byte to be received
	uint8_t byte = 0x80;
	// could do initialisations before timer reset to don't worry about WR line timing variations

	// receive the 7 following bits on a 4us basis
	while(bit!=0 && SP_IF_ENABLED)
	{
		// waiting next bit...
		while(TCNT1L == last);
		// update the last timer value
		last = TCNT1L;
		// sense the WR line
		uint8_t tmp1 = (SP_IWRD_PIN & SP_IWRD);
		// test if toggled
		if(tmp1 != line)
		{
			// the WR line toggled, the current bit is to be set
			line ^= SP_IWRD;
			byte |= bit;
		}
		bit = bit >> 1;
	}
	// keep the state of the WR line for the next byte
	g_sp_state.receive_line = line;
	return byte;
}

/**
called at the end of the receive isr
*/
void sp_receive_isr_end()
{
	// select the default bank
	CSET_BANK_UNSET;
	
	g_sp_state.receive_step = SP_END;
		
	// stop the timer...not required here
	//TCCR1B = 0;

	// disable our interrupt to don't be called after isr exit
	TIMSK1 = 0; //&= !(1<<ICIE1);
//die(LEDMASK_POWER,LEDMASK_WARNING,LEDMASK_ERROR,250);
	LEDs_TurnOffLEDs(LEDMASK_SP);
}
