/**
=============================================================================
Smartport Virtual Hard Drive, version 1.00.00
smartport (aka Cbus) protocol interpreter

If you would like to be involved in this project, or simply get my 
documentation for your own smartport project, feel free to contact me at
pcedric@hotmail.com

Copyright 2010,2011,2012 Cdric PELTIER, pcedric@hotmail.com

This file is part of the Smartport VHD project

Smartport VHD is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 3

Smartport VHD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Smartport VHD. If not, see <http://www.gnu.org/licenses/>.
=============================================================================
*/


#include "spvhd.h"
#include "protocol.h"
#include "cache.h"
#include <string.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "extsram.h"
#include "usb_storage.h"
#ifdef CFG_DEBUG_USB
#include "usb_serial.h"
#endif

/**
Called at the first incoming bit from the host
Receive a full packet, then let the application continue
This is required as the AVR is not power enough to manage
the smartport asynchronously
*/
ISR(TIMER1_CAPT_vect)
#if RECEIVE_ISR_V2 == 0
{
	// initial version
   // interrupts are disabled globaly by the interrupt call, 
	// no need to disabled our interrupt now
	// start of the critical timing phase
	// 44 cycles are lost by the interrupt call,
	// we have only 64 cycles between bits @16MHz,
	
	struct sp_receive_isr_low l_isr_state;
	
	// get the current state of the line to detect incoming bits
	l_isr_state.line = (SP_IWRD_PIN & SP_IWRD);

    SP_DEBUG_PUSH('+');

	// update visual indicator of smartport activity
	LEDs_TurnOnLEDs(LEDMASK_SP);	

	// receive synchronisation sequence and wait the start of the packet
	while(sp_receive_isr_get_byte(&l_isr_state)!=SP_PKTBEGIN)
	{
		//SP_DEBUG_PUSH(';');
		// check for smartport presence
		if(!SP_IF_ENABLED)
		{
			sp_receive_isr_end();
			g_sp_state.error = SP_ERRRESET;
			SP_DEBUG_PUSH('9');
			return;
		}
	}
	
	// receive header
	uint8_t* p_to = g_sp_state.p_buffer;
	uint8_t* p_packet = g_sp_state.p_packet;
	uint8_t* p_to_end = g_sp_state.p_buffer_end;
	uint8_t crc_computed = 0;
	
	//SP_DEBUG_PUSH(':');
	for(uint8_t counter=7;counter>0;counter--)
	{
		uint8_t byte = sp_receive_isr_get_byte(&l_isr_state);
		// check for smartport presence
		if(!SP_IF_ENABLED)
		{
			sp_receive_isr_end();
			g_sp_state.error = SP_ERRRESET;
			SP_DEBUG_PUSH('9');
			return;
		}
		//SP_DEBUG_PUSH('9');
		// update the computed crc and the packet header
		crc_computed ^= byte;
		byte &= (uint8_t)127; 
		*p_packet++ = byte;
	}
	
	// select the bank of memory used for caching blocks
	CSET_BANK_SET;

	// receive odds bytes
	if(g_sp_packet.header.nbodd>0)
	{
		uint8_t msb = sp_receive_isr_get_byte(&l_isr_state);
		//SP_DEBUG_PUSH('1');
		for(uint8_t counter=g_sp_packet.header.nbodd;counter>0;counter--)
		{
			//SP_DEBUG_PUSH('2');
			uint8_t byte = sp_receive_isr_get_byte(&l_isr_state);
			// check for smartport presence
			if(!SP_IF_ENABLED)
			{
				sp_receive_isr_end();
				g_sp_state.error = SP_ERRRESET;
				SP_DEBUG_PUSH('9');
				return;
			}
			msb = (msb<<1);
			byte = (byte&(uint8_t)(127))|(msb&(uint8_t)(128));
			crc_computed ^= byte;
			*p_to++ = byte;
		}
	}
	
	// receive group of seven bytes encoded as height bytes
	for(uint8_t counter2=g_sp_packet.header.nbgrp;counter2>0;counter2--)
	{
		// some group to receive
		uint8_t msb = sp_receive_isr_get_byte(&l_isr_state);
		// check for smartport presence
		if(!SP_IF_ENABLED)
		{
			sp_receive_isr_end();
			g_sp_state.error = SP_ERRRESET;
			SP_DEBUG_PUSH('9');
			return;
		}
		//SP_DEBUG_PUSH('3');
		for(uint8_t counter=7;counter>0;counter--)
		{
			//SP_DEBUG_PUSH('4');
			uint8_t byte = sp_receive_isr_get_byte(&l_isr_state);
			msb = (msb<<1);
			byte = (byte&(uint8_t)(127))|(msb&(uint8_t)(128));
			crc_computed ^= byte;
			*p_to = byte;
			// in this loop, we can trash memory over the end of the buffer
			// so better to check it
			if(p_to<p_to_end)
				++p_to;
		}
	}
	
	// select the default bank
	// all data are now stored
	CSET_BANK_UNSET;

	// receive first crc byte
	uint8_t tmp_crc1 = sp_receive_isr_get_byte(&l_isr_state);
	// check for smartport presence
	if(!SP_IF_ENABLED)
	{
		sp_receive_isr_end();
		g_sp_state.error = SP_ERRRESET;
		SP_DEBUG_PUSH('9');
		return;
	}
	*p_packet++ = tmp_crc1;
	
	// receive last crc byte
	uint8_t tmp_crc2 = sp_receive_isr_get_byte(&l_isr_state);
	// check for smartport presence
	if(!SP_IF_ENABLED)
	{
		sp_receive_isr_end();
		g_sp_state.error = SP_ERRRESET;
		SP_DEBUG_PUSH('9');
		return;
	}
	*p_packet++ = tmp_crc2;
	tmp_crc2 = tmp_crc2 << 1;
	tmp_crc2 |= 1;
	uint8_t crc_received = tmp_crc1 & tmp_crc2;
	// receive end packet mark
	tmp_crc1 = sp_receive_isr_get_byte(&l_isr_state);
	*p_packet++ = tmp_crc1;
	// receive the dummy byte
	tmp_crc2 = sp_receive_isr_get_byte(&l_isr_state);
	*p_packet++ = tmp_crc2;
	
	if(crc_received != crc_computed)
	{
		g_sp_state.error = SP_ERRCRC;
		SP_DEBUG_PUSH('-');
	}
	else if(tmp_crc1 != SP_PKTEND)
	{
		g_sp_state.error = SP_ERRPACKET;
	}
	else
	{
		// start the closing handshake now to avoid possible synchronization problem
		// only if the packet is for us
		struct sp_states* p_state = &g_sp_state; 
		uint8_t iddst = g_sp_packet.header.iddst;
		uint8_t drive_set = p_state->drive_set;
		uint8_t drive_max = p_state->drive_max;
		uint8_t drive_current = 255;
	
		// find the destination drive
		for(uint8_t i=0;i<drive_set && drive_current==255;i++)
			if(p_state->id[i]==iddst)
				drive_current = i;

		p_state->drive_current = drive_current;
		CSET_BANK_SET;
		uint8_t command = p_state->p_buffer[0];
		CSET_BANK_UNSET;
		if( (drive_current!=255 || (drive_set<drive_max && command==SP_CMDSID)) && g_sp_state.error==SP_ERROK )
		{
			SP_DO_OACKOFF;
		}
		if(drive_current == 255)
		{
			#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
			struct sp_packet* p_packet = &g_sp_packet;
			printf_P(PSTR("SPB: %02X->%02X <%02X %02X><%02X><%02X %02X><%02X %02X %02X %02X %02X %02X %02X %02X>\r\n"),p_packet->header.idsrc,iddst,p_packet->header.pkttype,p_packet->header.pktauxtype,p_packet->header.status,p_packet->header.nbodd,p_packet->header.nbgrp,command,p_state->p_buffer[1],p_state->p_buffer[2],p_state->p_buffer[3],p_state->p_buffer[4],p_state->p_buffer[5],p_state->p_buffer[6],p_state->p_buffer[7]);		
			#endif
		}
	}
	// end of the critical timing phase
	
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(crc_received != crc_computed)
	{
		printf_P(PSTR("SPB: crc error\r\n"));
		sp_debug_packet();
		#if defined(CFG_DEBUG_USB)
		usb_main();
		#endif
	}
	else if(tmp_crc1 != SP_PKTEND)
	{
		printf_P(PSTR("SPB: end packet error <%02X>\r\n"),tmp_crc1);
		sp_debug_packet();
		#if defined(CFG_DEBUG_USB)
		usb_main();
		#endif
	}
	#endif

	sp_receive_isr_end();
}
#else
{
	// 2nd version for the fast mode
   // interrupts are disabled globaly by the interrupt call, 
	// no need to disabled our interrupt now
	// start of the critical timing phase
	// 44 cycles are lost by the interrupt call,
	// we have only 64 cycles between bits @16MHz,
	
    SP_DEBUG_PUSH('+');
	// structure to decode the WR line from the host
	struct sp_receive_isr_low l_isr_state;
	// get the current state of the line to detect incoming bits
	l_isr_state.line = (SP_IWRD_PIN & SP_IWRD);
	// pointer used to stored the data
	uint8_t* p_to;
	// pointer to the packet structure
	uint8_t* p_packet;
	// pointer to the end of the buffer
	uint8_t* p_to_end;
	uint8_t l_crc_computed;
	// last byte to be stored
	uint8_t l_byte;
	// msb byte for the current group of data byte
	uint8_t l_msb;
	// destination of the last byte to be stored
	// 0 = packet structure
	// 1 = data buffer
	uint8_t l_byte_type;
	
	// update visual indicator of smartport activity
	LEDs_TurnOnLEDs(LEDMASK_SP);	

	// receive synchronisation sequence and wait the start of the packet
	// initialize pointers to store data
	l_isr_state.byte = 0xFF;
	while(l_isr_state.byte!=SP_PKTBEGIN)
	{
		sp_receive_isr_wait_msb(&l_isr_state);
		p_to = g_sp_state.p_buffer;
		sp_receive_isr_next_bit(&l_isr_state);
		p_packet = g_sp_state.p_packet;
		sp_receive_isr_next_bit(&l_isr_state);
		p_to_end = g_sp_state.p_buffer_end;
		sp_receive_isr_next_bit(&l_isr_state);
		l_crc_computed = 0;
		sp_receive_isr_next_bit(&l_isr_state);
		sp_receive_isr_next_bit(&l_isr_state);
		sp_receive_isr_last_bit(&l_isr_state);

		// check for smartport presence
		if(!SP_IF_ENABLED)
		{
			sp_receive_isr_end();
			g_sp_state.error = SP_ERRRESET;
			SP_DEBUG_PUSH('9');
			return;
		}
	}
	
	// receive header
	//SP_DEBUG_PUSH(':');
	// receive the first byte
	l_byte = sp_receive_isr_get_byte(&l_isr_state);
	for(uint8_t counter=6;counter>0;counter--)
	{
		// start to receive the current byte
		sp_receive_isr_wait_msb(&l_isr_state);
		// update the crc with the previous byte
		l_crc_computed ^= l_byte;
		sp_receive_isr_next_bit(&l_isr_state);
		// prepare the previous byte to be stored
		l_byte &= (uint8_t)127; 
		sp_receive_isr_next_bit(&l_isr_state);
		// store the previous byte
		*p_packet++ = l_byte;
		sp_receive_isr_next_bit(&l_isr_state);
		// the last header byte will be stored into the packet structure
		l_byte_type = 0;
		// finish the reception of the current byte
		sp_receive_isr_next_bit(&l_isr_state);
		sp_receive_isr_next_bit(&l_isr_state);
		l_byte = sp_receive_isr_last_bit(&l_isr_state);

		// check for smartport presence
		if(!SP_IF_ENABLED)
		{
			sp_receive_isr_end();
			g_sp_state.error = SP_ERRRESET;
			SP_DEBUG_PUSH('9');
			return;
		}
	}
	// the last header byte have to be stored
	
	// receive odds bytes
	if(g_sp_packet.header.nbodd>0)
	{
		// start the reception of the msb byte
		// and work with the last header's byte at the same time
		sp_receive_isr_wait_msb(&l_isr_state);
		l_crc_computed ^= l_byte;
		sp_receive_isr_next_bit(&l_isr_state);
		l_byte &= (uint8_t)127; 
		sp_receive_isr_next_bit(&l_isr_state);
		*p_packet++ = l_byte;
		sp_receive_isr_next_bit(&l_isr_state);
		// the last odd byte will be stored into the data buffer
		l_byte_type = 1;
		sp_receive_isr_next_bit(&l_isr_state);
		// select the bank of memory used for caching blocks
		CSET_BANK_SET;
		sp_receive_isr_next_bit(&l_isr_state);
		// keep the msb byte locally to extract its bits
		l_msb = sp_receive_isr_last_bit(&l_isr_state);
		
		//SP_DEBUG_PUSH('1');
		// receive the first odd byte
		l_byte = sp_receive_isr_get_byte(&l_isr_state);
		for(uint8_t counter=g_sp_packet.header.nbodd-1;counter>0;counter--)
		{
			//SP_DEBUG_PUSH('2');
			// start the reception of the current odd byte
			// and store the previous odd byte at the same time
			sp_receive_isr_wait_msb(&l_isr_state);
			l_msb = (l_msb<<1);
			sp_receive_isr_next_bit(&l_isr_state);
			l_byte = (l_byte&(uint8_t)(127))|(l_msb&(uint8_t)(128));
			sp_receive_isr_next_bit(&l_isr_state);
			l_crc_computed ^= l_byte;
			sp_receive_isr_next_bit(&l_isr_state);
			*p_to++ = l_byte;
			sp_receive_isr_next_bit(&l_isr_state);
			sp_receive_isr_next_bit(&l_isr_state);
			l_byte = sp_receive_isr_last_bit(&l_isr_state);

			// check for smartport presence
			if(!SP_IF_ENABLED)
			{
				sp_receive_isr_end();
				g_sp_state.error = SP_ERRRESET;
				SP_DEBUG_PUSH('9');
				return;
			}
		}
		// the last odd byte have to be stored
	}
	
	// receive group of seven bytes encoded as height bytes
	for(uint8_t l_counter2=g_sp_packet.header.nbgrp;l_counter2>0;l_counter2--)
	{
		// receive the msb of the next group
		if(l_byte_type==0)
		{
			// the last byte have to be stored into the packet structure
			// start the reception of the msb byte
			// and work with the last header's byte at the same time
			sp_receive_isr_wait_msb(&l_isr_state);
			// select the usual bank of memory
			CSET_BANK_UNSET;
			sp_receive_isr_next_bit(&l_isr_state);
			l_crc_computed ^= l_byte;
			sp_receive_isr_next_bit(&l_isr_state);
			l_byte &= (uint8_t)127; 
			sp_receive_isr_next_bit(&l_isr_state);
			*p_packet++ = l_byte;
			sp_receive_isr_next_bit(&l_isr_state);
		}
		else
		{
			// the last byte have to be stored into the data buffer
			// start the reception of the msb byte
			// and work with the last byte at the same time
			sp_receive_isr_wait_msb(&l_isr_state);
			// select the bank of memory used for caching blocks
			CSET_BANK_SET;
			sp_receive_isr_next_bit(&l_isr_state);
			l_msb = (l_msb<<1);
			sp_receive_isr_next_bit(&l_isr_state);
			l_byte = (l_byte&(uint8_t)(127))|(l_msb&(uint8_t)(128));
			sp_receive_isr_next_bit(&l_isr_state);
			l_crc_computed ^= l_byte;
			sp_receive_isr_next_bit(&l_isr_state);
			*p_to++ = l_byte;
		}
		sp_receive_isr_next_bit(&l_isr_state);
		// now, we are working with data byte only
		l_byte_type=1;
		l_msb = sp_receive_isr_last_bit(&l_isr_state);

		// check for smartport presence
		if(!SP_IF_ENABLED)
		{
			sp_receive_isr_end();
			g_sp_state.error = SP_ERRRESET;
			SP_DEBUG_PUSH('9');
			return;
		}
		//SP_DEBUG_PUSH('3');
		// receive the first byte of the group
		l_byte = sp_receive_isr_get_byte(&l_isr_state);
		for(uint8_t l_counter=6;l_counter>0;l_counter--)
		{
			//SP_DEBUG_PUSH('4');
			// start the reception of the next byte
			// and store the previous byte at the same time
			sp_receive_isr_wait_msb(&l_isr_state);
			l_msb = (l_msb<<1);
			sp_receive_isr_next_bit(&l_isr_state);
			l_byte = (l_byte&(uint8_t)(127))|(l_msb&(uint8_t)(128));
			sp_receive_isr_next_bit(&l_isr_state);
			l_crc_computed ^= l_byte;
			sp_receive_isr_next_bit(&l_isr_state);
			*p_to = l_byte;
			sp_receive_isr_next_bit(&l_isr_state);
			// in this loop, we can trash memory over the end of the buffer
			// so better to check it
			if(p_to<p_to_end)
				++p_to;
			sp_receive_isr_next_bit(&l_isr_state);
			l_byte = sp_receive_isr_last_bit(&l_isr_state);
		}
		// the last byte have to be stored
		// it will be stored during the reception
		// of the msb byte of the next group
	}

	// receive first crc byte
	if(l_byte_type==0)
	{
		// the last byte have to be stored into the packet structure
		// start the reception of the crc byte
		// and work with the last header's byte at the same time
		sp_receive_isr_wait_msb(&l_isr_state);
		l_crc_computed ^= l_byte;
		sp_receive_isr_next_bit(&l_isr_state);
		l_byte &= (uint8_t)127; 
		sp_receive_isr_next_bit(&l_isr_state);
		CSET_BANK_UNSET;
		sp_receive_isr_next_bit(&l_isr_state);
		*p_packet++ = l_byte;
		sp_receive_isr_next_bit(&l_isr_state);
	}
	else
	{
		// the last byte have to be stored into the data buffer
		// start the reception of the crc byte
		// and work with the last byte at the same time
		sp_receive_isr_wait_msb(&l_isr_state);
		l_msb = (l_msb<<1);
		sp_receive_isr_next_bit(&l_isr_state);
		l_byte = (l_byte&(uint8_t)(127))|(l_msb&(uint8_t)(128));
		sp_receive_isr_next_bit(&l_isr_state);
		l_crc_computed ^= l_byte;
		sp_receive_isr_next_bit(&l_isr_state);
		CSET_BANK_SET;
		sp_receive_isr_next_bit(&l_isr_state);
		*p_to++ = l_byte;
	}
	sp_receive_isr_next_bit(&l_isr_state);
	// select the default bank, all data are now stored
	CSET_BANK_UNSET;
	uint8_t l_tmp_crc1 = sp_receive_isr_last_bit(&l_isr_state);
	
	// check for smartport presence
	if(!SP_IF_ENABLED)
	{
		sp_receive_isr_end();
		g_sp_state.error = SP_ERRRESET;
		SP_DEBUG_PUSH('9');
		return;
	}
	
	// receive last crc byte and store the first crc byte
	sp_receive_isr_wait_msb(&l_isr_state);
	*p_packet = l_tmp_crc1;
	sp_receive_isr_next_bit(&l_isr_state);
	p_packet++;
	sp_receive_isr_next_bit(&l_isr_state);
	sp_receive_isr_next_bit(&l_isr_state);
	sp_receive_isr_next_bit(&l_isr_state);
	sp_receive_isr_next_bit(&l_isr_state);
	uint8_t l_tmp_crc2 = sp_receive_isr_last_bit(&l_isr_state);
	
	// check for smartport presence
	if(!SP_IF_ENABLED)
	{
		sp_receive_isr_end();
		g_sp_state.error = SP_ERRRESET;
		SP_DEBUG_PUSH('9');
		return;
	}

	// receive the end packet mark and work on the crc
	sp_receive_isr_wait_msb(&l_isr_state);
	*p_packet = l_tmp_crc2;
	sp_receive_isr_next_bit(&l_isr_state);
	p_packet++;
	sp_receive_isr_next_bit(&l_isr_state);
	l_tmp_crc2 = l_tmp_crc2 << 1;
	l_tmp_crc2 |= 1;
	sp_receive_isr_next_bit(&l_isr_state);
	uint8_t l_crc_received = l_tmp_crc1 & l_tmp_crc2;
	sp_receive_isr_next_bit(&l_isr_state);
	sp_receive_isr_next_bit(&l_isr_state);
	l_tmp_crc1 = sp_receive_isr_last_bit(&l_isr_state);

	// receive the dummy byte and start final task
	sp_receive_isr_wait_msb(&l_isr_state);
	*p_packet = l_tmp_crc1;
	sp_receive_isr_next_bit(&l_isr_state);

	// in case of error, simply ignore the remaining bits of the dummy byte		
	if(l_crc_received != l_crc_computed)
	{
		g_sp_state.error = SP_ERRCRC;
		SP_DEBUG_PUSH('-');
	}
	else if(l_tmp_crc1 != SP_PKTEND)
	{
		g_sp_state.error = SP_ERRPACKET;
	}
	else
	{
		sp_receive_isr_next_bit(&l_isr_state);
		struct sp_states* p_state = &g_sp_state; 
		sp_receive_isr_next_bit(&l_isr_state);
		uint8_t l_iddst = g_sp_packet.header.iddst;
		sp_receive_isr_next_bit(&l_isr_state);
		uint8_t l_drive_set = p_state->drive_set;
		sp_receive_isr_next_bit(&l_isr_state);
		uint8_t l_drive_max = p_state->drive_max;
		// the dummy byte is simply discarded
		sp_receive_isr_last_bit(&l_isr_state);
		uint8_t l_drive_current = 255;
	
		// find the destination drive
		for(uint8_t i=0;i<l_drive_set && l_drive_current==255;i++)
			if(p_state->id[i]==l_iddst)
				l_drive_current = i;

		p_state->drive_current = l_drive_current;
		CSET_BANK_SET;
		uint8_t l_command = p_state->p_buffer[0];
		CSET_BANK_UNSET;
		// start the closing handshake now to avoid possible synchronization problem
		// but only if the packet is for us
		if( (l_drive_current!=255 || (l_drive_set<l_drive_max && l_command==SP_CMDSID)) && g_sp_state.error==SP_ERROK )
		{
			SP_DO_OACKOFF;
		}
		#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
		if(l_drive_current == 255)
		{
			struct sp_packet* p_packet = &g_sp_packet;
			printf_P(PSTR("SPB: %02X->%02X <%02X %02X><%02X><%02X %02X><%02X %02X %02X %02X %02X %02X %02X %02X>\r\n"),p_packet->header.idsrc,l_iddst,p_packet->header.pkttype,p_packet->header.pktauxtype,p_packet->header.status,p_packet->header.nbodd,p_packet->header.nbgrp,l_command,p_state->p_buffer[1],p_state->p_buffer[2],p_state->p_buffer[3],p_state->p_buffer[4],p_state->p_buffer[5],p_state->p_buffer[6],p_state->p_buffer[7]);		
		}
		#endif
	}
	// end of the critical timing phase
	
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(l_crc_received != l_crc_computed)
	{
		printf_P(PSTR("SPB: crc error\r\n"));
		sp_debug_packet();
		#if defined(CFG_DEBUG_USB)
		usb_main();
		#endif
	}
	else if(l_tmp_crc1 != SP_PKTEND)
	{
		printf_P(PSTR("SPB: end packet error <%02X>\r\n"),l_tmp_crc1);
		sp_debug_packet();
		#if defined(CFG_DEBUG_USB)
		usb_main();
		#endif
	}
	#endif

	sp_receive_isr_end();
}
#endif

/**
Read a byte from the smartport synchronously
Timer1 Input capture is used to synchronize reception on the msb
Timer1 /64 is then used to catch the following 7 bits on a 4us basis
*/
uint8_t sp_receive_isr_get_byte(struct sp_receive_isr_low* lp_isr_state)
{
	// waiting msb to synchronize the reception...
	uint8_t line = lp_isr_state->line;
	// force the loop statement to be 8 bits computing
	uint8_t tmp;
	while( (tmp=(SP_IWRD_PIN & SP_IWRD)) == line )
	{
		// check for smartport presence
		if(!SP_IF_ENABLED)
			return 0;
	}
	// reset the prescaler so the first increment will be a full period
	GTCCR = (1<<PSRSYNC);
	// reset timer value
	TCNT1L = 0;
	// update the line state
	line ^= SP_IWRD;
	// set timer value to detect timer change
	uint8_t last = 0;
	// current bit to be received
	uint8_t bit = 0x40;
	// current byte to be received
	uint8_t byte = 0x80;
	// could do initialisations before timer reset to don't worry about WR line timing variations

	// receive the 7 following bits on a 4us basis
	while(bit!=0 && SP_IF_ENABLED)
	{
		// waiting next bit...
		while(TCNT1L == last);
		// update the last timer value
		last = TCNT1L;
		// sense the WR line
		uint8_t tmp1 = (SP_IWRD_PIN & SP_IWRD);
		// test if toggled
		if(tmp1 != line)
		{
			// the WR line toggled, the current bit is to be set
			line ^= SP_IWRD;
			byte |= bit;
		}
		bit = bit >> 1;
	}
	// keep the state of the WR line for the next byte
	lp_isr_state->line = line;
	return byte;
}

/**
Wait for the reception of the next msb and read bit 6
Timer1 Input capture is used to synchronize reception on the msb
Timer1 /64 is then used to catch the following 7 bits on a 4us basis
*/
void sp_receive_isr_wait_msb(struct sp_receive_isr_low* lp_isr_state)
{
	uint8_t line = lp_isr_state->line;
	// force the loop statement to be 8 bits computing
	uint8_t tmp;
	while( (tmp=(SP_IWRD_PIN & SP_IWRD)) == line )
	{
		// check for smartport presence
		if(!SP_IF_ENABLED)
			return;
	}
	// reset the prescaler so the first increment will be a full period
	GTCCR = (1<<PSRSYNC);
	// reset timer value
	TCNT1L = 0;
	// update the line state
	lp_isr_state->line ^= SP_IWRD;
	// set timer value to detect timer change
	lp_isr_state->last = 0;
	// current bit to be received
	lp_isr_state->bit = 0x40;
	// current byte to be received
	lp_isr_state->byte = 0x80;
	// could do initialisations before timer reset to don't worry about 
	// WR line timing variations
	sp_receive_isr_next_bit(lp_isr_state);
}

/**
Wait and read the next bit to be received
*/
void sp_receive_isr_next_bit(struct sp_receive_isr_low* lp_isr_state)
{
	// receive the following bits after 4us in slow mode
	while(TCNT1L == lp_isr_state->last);
	// update the last timer value
	lp_isr_state->last = TCNT1L;
	// sense the WR line
	uint8_t tmp1 = (SP_IWRD_PIN & SP_IWRD);
	// test if toggled
	if(tmp1 != lp_isr_state->line)
	{
		// the WR line toggled, the current bit is to be set
		lp_isr_state->line ^= SP_IWRD;
		lp_isr_state->byte |= lp_isr_state->bit;
	}
	lp_isr_state->bit = lp_isr_state->bit >> 1;
}

/**
Wait and read the last bit to be received
*/
uint8_t sp_receive_isr_last_bit(struct sp_receive_isr_low* lp_isr_state)
{
	// receive the following bits after 4us in slow mode
	while(TCNT1L == lp_isr_state->last);
	// update the last timer value
	lp_isr_state->last = TCNT1L;
	// sense the WR line
	uint8_t tmp1 = (SP_IWRD_PIN & SP_IWRD);
	uint8_t byte = lp_isr_state->byte;
	// test if toggled
	if(tmp1 != lp_isr_state->line)
	{
		// the WR line toggled, the current bit is to be set
		lp_isr_state->line ^= SP_IWRD;
		byte |= lp_isr_state->bit;
	}
	return byte;
}

/**
called at the end of the receive isr
*/
void sp_receive_isr_end()
{
	// select the default bank
	CSET_BANK_UNSET;
	
	g_sp_state.receive_step = SP_END;
		
	// stop the timer...not required here
	//TCCR1B = 0;

	// disable our interrupt to don't be called after isr exit
	TIMSK1 = 0; //&= !(1<<ICIE1);
//die(LEDMASK_POWER,LEDMASK_WARNING,LEDMASK_ERROR,250);
	LEDs_TurnOffLEDs(LEDMASK_SP);
}
