/**
=============================================================================
Smartport Virtual Hard Drive, version 0.99.0 (beta)
smartport (aka Cbus) protocol interpreter

If you would like to be involved in this project, or simply get my 
documentation for your own smartport project, feel free to contact me at
pcedric@hotmail.com

Copyright 2010 Cédric PELTIER, pcedric@hotmail.com

This file is part of the Smartport VHD project

Smartport VHD is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 3

Smartport VHD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Smartport VHD. If not, see <http://www.gnu.org/licenses/>.
=============================================================================
*/


#include "spvhd.h"
#include "protocol.h"
#include "cache.h"
#include "ds1307.h"
#include <avr/pgmspace.h>
#include <string.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "extsram.h"
#ifdef CFG_DEBUG_USB
#include "usb_serial.h"
#endif

/// Device Information Block description
/// Byte 0, general status byte
///  >7 =1 for block device =0 for character device
///  >6 =1 write allowed
///  >5 =1 read allowed
///  >4 =1 device online or disk in drive
///  >3 =1 format allowed
///   2 =1 media is write protected (block device only)
///   1 =1 device currently interrupting (//c only)
///   0 =1 device currently open (character device only)
/// Byte 1 (low), 2 (medium), 3 (high), size of the device in block, all zero for character device
/// Byte 4, lenght of the id string
/// Byte 5 to 20, id string
/// Byte 21, device type
///  =00 ram card
///  =01 3.5 floppy disk
///  =02 Profile type hard disk
///  =03 generic scsi device
///  =04 rom disk
///  =05 scsi cd rom
///  =06 scsi tape or sequential
///  =07 scsi hard disk
///  =08 reserved
///  =09 scsi printer
///  =0A 5.25 floppy drive
///  =0B reserved
///  =0C reserved
///  =0D printer
///  =0E clock (but Prodos 8 never ask date and time to it, tested)
///  =0F modem
/// Byte 22, device sub type
///  7 =1 support extended smartport
///  6 =1 support disk switched errors
///  5 =0 for removable media
///  4 reserved
///  3 reserved
///  2 reserved
///  1 reserved
///  0 reserved
/// Byte 23, 24, version word
const uint8_t const g_sp_dib_template[SP_DIBSIZE] PROGMEM = {0xF8,0,0,0,5,'D','r','i','v','e',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',0x02,0x20,0x0B,0x10};

/// Device control block of the smartportvhd
const uint8_t const g_sp_dcb[] PROGMEM = {1,0,0,0,SP_MAX_DRIVE,5,'S','P','V','H','D'};

/**
Be ready to send the current packet
@param type Type of the packet to write in the header
@param status Status byte to write in the header
@param p_buffer POinter to the data buffer
@param size Size of the data buffer
*/
void sp_send_init(uint8_t type,uint8_t status,uint8_t* p_buffer,uint16_t size)
{
    SP_DEBUG_PUSH('s');

    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug & CFG_DEBUG_MASK_SND)        printf_P(PSTR("SPB: sp_send_init()\r\n"));
    #endif

    cli();
    
    // disable interrupt for now
    TIMSK1=0;
    // clearing pending interrupt
    TIFR1=(1<<OCIE1C)|(1<<ICF1);
    // smartport cycles of 4us (@16MHz)
    ICR1 =63;
    // pulse width of 1us (@16MHz)
    OCR1C=15;
    // Set the counter 1 start value
    TCNT1 = 0;
    // fast pwm mode with top = ICR1
    // set OC1C on top, clear OC1C on compare match, this give positives pulses TCCR1A=(1<<COM1C1)
    // clear OC1C on top, set on compare match, this give negatives pulses TCCR1A=(1<<COM1C1)|(1<<COM1C0)
    // but don't start the counter now, wait for the host to be ready
    // disconnect OC0A which is the same pin as OC1C
    // connect OC1C pin using switch COM1C1 and COM1C0
    TCCR0A=0;//(TCCR0A&(255-(1<<WGM00)-(1<<WGM01)));
    TCCR1A=(1<<COM1C1)|(1<<COM1C0)|(1<<WGM11);
    TCCR1B=(1<<WGM13)|(1<<WGM12);// clock stopped for  now |(1<<CS10);

    sei();
    
    sp_init_packet(1, type, status, p_buffer, size);
}

/**
Encode the data into a packet to be sent, update crc with encoded data
write the crc, the end of packet mark, and the dummy byte
g_sp_state.size_computed must contain the data size to send
@return enum sp_error
*/
enum sp_error sp_encode()
#if RECEIVE_ISR_V2 == 0
{
	struct sp_packet* p_packet = &g_sp_packet;
	struct sp_states* p_state = &g_sp_state;
    uint8_t* p_from = p_state->p_buffer;
    uint8_t* p_to = p_packet->data;

    uint8_t crc = 0;
    // compute the number of group and odd bytes 
    uint16_t nbgrp = p_state->size_computed/7;
    uint16_t nbodd = p_state->size_computed-(nbgrp*7);
    gp_sp_sisr_end = p_to + nbodd + (nbodd==0?0:1) + (nbgrp * 8) + 3;  
	if(gp_sp_sisr_end>(uint8_t*)(&p_packet->canary))
		return SP_ERROVERFLOW;
    
    // update the header of the packet to be sent
    p_packet->header.nbodd=(uint8_t)(nbodd)|128;
    p_packet->header.nbgrp=(uint8_t)(nbgrp)|128;

    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    // Could be too long most of the time
    if(gb_debug & CFG_DEBUG_MASK_SND)        printf_P(PSTR("SPB: sp_encode() from %04X size %d, %d odds, %d groups\r\n"),p_from,p_state->size_computed,nbodd,nbgrp);
    //printf_P(PSTR("SPB: %02X %02X %02X %02X\r\n"),p_from[0],p_from[1],p_from[2],p_from[3]);
    #endif

    #ifdef CFG_MEMSET
    memset(&p_packet->data,0,(&p_packet->canary-&p_packet->data));
    #endif
    
    // start to send synchronization sequence and header will encoding the data
    // this is time critical and dependant of the speed of the processor
    sp_send_start();
    
    // read data to be encoded from the right bank of memory
    //if(p_from >= (uint8_t*)(EXT_SRAM_START))
    CSET_BANK_SET;

    if(nbodd>0)
    {
        // there is some odd bytes to encode 
        uint8_t msb = 1;
        uint8_t* p = p_to;
        for(uint8_t i=nbodd;i!=0;i--)
        {
            uint8_t byte = *p_from++; 
            crc ^= byte;
            msb = msb<<1;
            if(byte>(uint8_t)(127))
                msb|=1;
            byte |= 128;
            *(++p_to) = byte;
        }
        *p = (msb<<(7-nbodd));  
        p_to++;
    }

    for(uint8_t j=nbgrp;j!=0;j--)
    {
        // encode the next group
        uint8_t msb = 1;
        uint8_t* p = p_to;
        for(uint8_t i=7;i!=0;i--)
        {
            uint8_t byte = *p_from++; 
            crc ^= byte;
            msb = msb<<1;
            if(byte>(uint8_t)(127))
                msb|=1;
            byte |= 128;
            *(++p_to) = byte;
        }
        *p = msb;  
        p_to++;
    }
    
    // use the default bank now, all data are encoded
  	CSET_BANK_UNSET;

    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    //printf_P(PSTR("SPB: crc %02X\r\n"),crc);
    #endif
    
    // update the crc
    for(uint8_t i=sizeof(struct sp_header)-1;i!=255;i--)
    {
    	uint8_t byte = ((uint8_t*)(&p_packet->header))[i]; 
        crc ^= byte;
    }
    // write the encoded crc
    *p_to = (crc | 0xAA);
    p_to++;
    crc = crc>>1;
    *p_to = (crc | 0xAA);
    p_to++;
    // add the end packet marker
    *p_to = SP_PKTEND;
    p_to++;
    // ...add a dummy byte, not required because we don't have a real iwm (?)
    *p_to = 0xCC;
//    gp_sp_sisr_end = p_to;//+1;

    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    // Could be too long most of the time
    if(gb_debug & CFG_DEBUG_MASK_SND)        printf_P(PSTR("SPB: sp_encode() end %04X\r\n"),gp_sp_sisr_end);
    #endif
    
    //p_state->crc_computed = crc;
    return SP_ERROK;
}
#else
{
	struct sp_packet* p_packet = &g_sp_packet;
	struct sp_states* p_state = &g_sp_state;
    uint8_t* p_from = p_state->p_buffer;
    uint8_t* p_to = p_packet->data;

    uint8_t crc = 0;
    // compute the number of group and odd bytes 
    uint16_t nbgrp = p_state->size_computed/7;
    uint16_t nbodd = p_state->size_computed-(nbgrp*7);
    gp_sp_sisr_end = p_to + nbodd + (nbodd==0?0:1) + (nbgrp * 8) + 3;  
	if(gp_sp_sisr_end>(uint8_t*)(&p_packet->canary))
		return SP_ERROVERFLOW;
    
    // update the header of the packet to be sent
    p_packet->header.nbodd=(uint8_t)(nbodd)|128;
    p_packet->header.nbgrp=(uint8_t)(nbgrp)|128;

    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    // Could be too long most of the time
    if(gb_debug & CFG_DEBUG_MASK_SND)        printf_P(PSTR("SPB: sp_encode() from %04X size %d, %d odds, %d groups\r\n"),p_from,p_state->size_computed,nbodd,nbgrp);
    //printf_P(PSTR("SPB: %02X %02X %02X %02X\r\n"),p_from[0],p_from[1],p_from[2],p_from[3]);
    #endif

    #ifdef CFG_MEMSET
    memset(&p_packet->data,0,(&p_packet->canary-&p_packet->data));
    #endif
    
    // start to send synchronization sequence and header will encoding the data
    // this is time critical and dependant of the speed of the processor
    sp_send_start();
    
    // read data to be encoded from the right bank of memory
    //if(p_from >= (uint8_t*)(EXT_SRAM_START))
    CSET_BANK_SET;

    if(nbodd>0)
    {
        // there is some odd bytes to encode 
        uint8_t msb = 1;
        uint8_t* p = p_to;
        for(uint8_t i=nbodd;i!=0;i--)
        {
            uint8_t byte = *p_from++; 
            crc ^= byte;
            msb = msb<<1;
            if(byte>(uint8_t)(127))
                msb|=1;
            byte |= 128;
            *(++p_to) = byte;
        }
        *p = (msb<<(7-nbodd));  
        p_to++;
    }

    for(uint8_t j=nbgrp;j!=0;j--)
    {
        // encode the next group
        uint8_t msb = 1;
        uint8_t* p = p_to;
        for(uint8_t i=7;i!=0;i--)
        {
            uint8_t byte = *p_from++; 
            crc ^= byte;
            msb = msb<<1;
            if(byte>(uint8_t)(127))
                msb|=1;
            byte |= 128;
            *(++p_to) = byte;
        }
        *p = msb;  
        p_to++;
    }
    
    // use the default bank now, all data are encoded
  	CSET_BANK_UNSET;

    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    //printf_P(PSTR("SPB: crc %02X\r\n"),crc);
    #endif
    
    // update the crc
    for(uint8_t i=sizeof(struct sp_header)-1;i!=255;i--)
    {
    	uint8_t byte = ((uint8_t*)(&p_packet->header))[i]; 
        crc ^= byte;
    }
    // write the encoded crc
    *p_to = (crc | 0xAA);
    p_to++;
    crc = crc>>1;
    *p_to = (crc | 0xAA);
    p_to++;
    // add the end packet marker
    *p_to = SP_PKTEND;
    p_to++;
    // ...add a dummy byte, not required because we don't have a real iwm (?)
    *p_to = 0xCC;
//    gp_sp_sisr_end = p_to;//+1;

    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    // Could be too long most of the time
    if(gb_debug & CFG_DEBUG_MASK_SND)        printf_P(PSTR("SPB: sp_encode() end %04X\r\n"),gp_sp_sisr_end);
    #endif
    
    //p_state->crc_computed = crc;
    return SP_ERROK;
}

/**
Initialize 
Timer1 Input capture is used to synchronize reception on the msb
Timer1 /64 is then used to catch the following 7 bits on a 4us basis
*/
void sp_send_msb(struct sp_send_low* lp_snd_state)
{
	// reset the prescaler so the first increment will be a full period
	GTCCR = (1<<PSRSYNC);
	// reset timer value
	TCNT1L = 0;
	// read the byte to be sent
	lp_isr_state->byte = *lp_isr_state->p_from;
	// move on to the next byte for the next time
	lp_isr_state->p_from++;
	// current bit to be sent
	lp_isr_state->bit = 0x80;
	sp_send_bit(lp_snd_state);
}
/**
Send the next bit
*/
void sp_send_next_bit(struct sp_send_low* lp_snd_state)
{
	// wait the end of the current pulse
	while( (TIFR1&(1<<OCF1C))==0 );
	
	// next bit
	lp_isr_state->bit = lp_isr_state->bit >> 1;
}
/**
Send the last bit
*/
void sp_send_last_bit(struct sp_send_low* lp_snd_state)
{
	// wait the end of the current pulse
	while( (TIFR1&(1<<OCF1C))==0 );
	
	// next bit
	lp_isr_state->bit = lp_isr_state->bit >> 1;
}

#endif

/**
Send the current packet now
*/
void sp_send_start()
{
    cli();
    SP_DO_ORDDENBL;
#if RECEIVE_ISR_V2 == 0
    TIMSK1=(1<<OCIE1C);//(1<<TOIE1);
#else
	// the 2nd version don't use isr
    TIMSK1=0;//(1<<TOIE1);
#endif
    TCCR1B|=(1<<CS10);
    sei();
    // debug only
    //PINB = (1<<PB0);
    //PINB = (1<<PB0);
    g_sp_state.send_step=SP_SYNC;
    SP_DEBUG_PUSH('S');
}

/**
Manage sending of data to the smartport
Once the handshake is done, activity go back to SP_NONE
*/
void sp_send_main()
{
    switch(g_sp_state.send_step)
    {
        case SP_WAIT:
			if(sp_handshake_opening()==0)
			{
				// send the current packet now
				// start the timer 1 with a direct clock and enable
				// interupt on top to control the following pulse
				_delay_us(30);
				LEDs_TurnOnLEDs(LEDMASK_SP);
//				sp_send_start();
			    switch(sp_encode())
				{
					case SP_ERROK:
						break;
					case SP_ERROVERFLOW:
						SP_DEBUG_PUSH('#');
						break;
					default:
						SP_DEBUG_PUSH('!');
						break;
				}
			}
			else
			{
				sp_receive_init();
				//LEDs_TurnOffLEDs(LEDMASK_SP);
			}
			break;

		case SP_SYNC:
			// the isr start the handshake when data are sent
			// wait the host handshake to go ahead
			#ifdef CFG_DEBUG_PUSH
			if(SP_IF_OACK)
				SP_DEBUG_PUSH('y');
			else
				SP_DEBUG_PUSH('a');
			#endif
	
			if(!SP_IF_REQUEST)
			{
				LEDs_TurnOffLEDs(LEDMASK_SP);
				
				#ifdef CFG_READ_AHEAD
				// read ahead cache system
				// after sending a data packet, read the next block from the current drive
				struct sp_states* p_state = &g_sp_state;
				if(p_state->block_address_next != 0xFFFFFFFF)
				{
					// find the block in the cache memory
					uint8_t* pointer = cset_get(p_state->drive_current,p_state->block_address_next,0);
					if(pointer == 0)
					{
						// not found in the cache, need to retrieve a buffer and load the block
						pointer = cset_new(p_state->drive_current,p_state->block_address_next,0);
						FIL* drive = &p_state->drive[p_state->drive_current];
						FRESULT rc = f_lseek(drive,(DWORD)(p_state->block_address_next)*(DWORD)(SP_BLOCKSIZE));
						if(rc==FR_OK)
						{
							UINT size_read=0;
							CSET_BANK_SET;			
							rc = f_read(drive,pointer,SP_BLOCKSIZE,&size_read);
							CSET_BANK_UNSET;			
						}
						if(rc!=FR_OK)
						{
							#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
							printf_P(PSTR("SPB: reading error %d\r\n"),rc);
							#endif
						}
					}
				}
				// avoid the read ahead overload until a new block is read
				p_state->block_address_next = 0xFFFFFFFF;
				#endif
	
				// after the read ahead to keep synchronisation with host if read ahead is too long
				// be ready to listen the host now
				sp_receive_init();

			}
			break;
    }
}

/**
send a device status
@param status Status to send to the host
*/
void sp_send_ds(uint8_t status)
{
    SP_DEBUG_PUSH('?');

    // use the start of the dib data as buffer to pack
    sp_send_init(SP_PKTDAT,status,&g_sp_state.dib[g_sp_state.drive_current][0],4);
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug & CFG_DEBUG_MASK_SND)        printf_P(PSTR("SPB: Send device status drive %d id %d\r\n"),g_sp_state.drive_current,g_sp_state.id[g_sp_state.drive_current]);
    #endif
}

/**
send a device information block
@param status Status to send to the host
*/
void sp_send_dib(uint8_t status)
{
    SP_DEBUG_PUSH('$');

    // use the start of the dib data as buffer to pack
    sp_send_init(SP_PKTDAT,status,&g_sp_state.dib[g_sp_state.drive_current][0],SP_DIBSIZE);
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug & CFG_DEBUG_MASK_SND)        printf_P(PSTR("SPB: Send device information block drive %d id %d\r\n"),g_sp_state.drive_current,g_sp_state.id[g_sp_state.drive_current]);
    #endif
    
/*    
	// requisite a buffer to send the dib
	struct spvhd_dcb* pointer = (struct spvhd_dcb*)cset_new(0xFF,0xFFFFFFFF,0);
	
	// Using the dcb template
	memcpy_P(pointer,g_sp_dcb,sizeof(g_sp_dcb));
	// Retreiving the lowest smartport id managed by the spvhd
	pointer->spid = sp_lowest_id_get(p_state);
	
	p_state->p_buffer = (uint8_t*)pointer;
	p_state->p_buffer_end = ((uint8_t*)(pointer))+sizeof(g_sp_dcb);

    sp_send_init(SP_PKTDAT,0,(uint8_t*)pointer,sizeof(g_sp_dcb));
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug & CFG_DEBUG_MASK_SND)        puts_P(PSTR("SPB: Send dcb FF\r\n"));
    #endif
*/
}

/**
send a status packet
@param status Status to send to the host
*/
void sp_send_status(uint8_t status)
{
    SP_DEBUG_PUSH('.');

    sp_send_init(SP_PKTSTA,status,g_sp_buffer,0);
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug & CFG_DEBUG_MASK_SND)        printf_P(PSTR("SPB: Send status %d drive %d id %d\r\n"),status,g_sp_state.drive_current,g_sp_state.id[g_sp_state.drive_current]);
    #endif
}

/**
send a data block
*/
void sp_send_data(uint8_t* data)
{
    SP_DEBUG_PUSH('&');

    sp_send_init(SP_PKTDAT,0,data,SP_BLOCKSIZE);
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug & CFG_DEBUG_MASK_SND)        printf_P(PSTR("SPB: Send block drive %d id %d\r\n"),g_sp_state.drive_current,g_sp_state.id[g_sp_state.drive_current]);
    #endif
}

/**
send the spvhd dcb
*/
void sp_send_dcb(struct sp_states* p_state)
{
    SP_DEBUG_PUSH('|');
    
	// requisite a buffer to send the dcb
	//struct spvhd_dcb* lp_spvhd_dcb = (struct spvhd_dcb*)cset_new(0xFF,0xFFFFFFFF,0);
	// recycle the end of the buffer used for nibbled data within the packet object
	struct spvhd_dcb* lp_spvhd_dcb = (struct spvhd_dcb*)&g_sp_packet.data[SP_BUFDATSIZE - sizeof(g_sp_dcb)]; 
	
	CSET_BANK_SET;
	// Using the dcb template
	memcpy_P(lp_spvhd_dcb,g_sp_dcb,sizeof(g_sp_dcb));
	// Retrieving the lowest smartport id managed by the spvhd
	lp_spvhd_dcb->spid = p_state->id[p_state->drive_current];//sp_lowest_id_get(p_state);
	CSET_BANK_UNSET;
	
	p_state->p_buffer = (uint8_t*)lp_spvhd_dcb;
	p_state->p_buffer_end = ((uint8_t*)(lp_spvhd_dcb))+sizeof(g_sp_dcb);

    sp_send_init(SP_PKTDAT,0,(uint8_t*)lp_spvhd_dcb,sizeof(g_sp_dcb));
    #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
    if(gb_debug & CFG_DEBUG_MASK_SND)        puts_P(PSTR("SPB: Send dcb\r\n"));
    #endif
}

/**
send the date
*/
void sp_send_date(struct sp_states* p_state)
{
    SP_DEBUG_PUSH('|');
	// recycle the end of the buffer used for nibbled data within the packet object
	uint8_t* lp_ptr = &g_sp_packet.data[SP_BUFDATSIZE - 4]; 
   bool l_rc;
	uint8_t l_data[9];
	uint8_t l_tmp;
	l_rc = ds1307_read(l_data);
	if(l_rc)
	{
		CSET_BANK_SET;
		// minutes
		lp_ptr[2] = bcd2bin(l_data[1]);
		// hours
		lp_ptr[3] = bcd2bin(l_data[2]) & 0x3F;
		// days and months
		l_tmp = bcd2bin(l_data[5]);
		lp_ptr[0] = (bcd2bin(l_data[4]) & 31) | (l_tmp<<5);
		// months and years
		lp_ptr[1] = (l_tmp>>3) | (bcd2bin(l_data[6])<<1);	
		CSET_BANK_UNSET;
		
		p_state->p_buffer = lp_ptr;
		p_state->p_buffer_end = lp_ptr+4;
	
		 sp_send_init(SP_PKTDAT,0,lp_ptr,4);
		 #if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
		 if(gb_debug & CFG_DEBUG_MASK_SND) puts_P(PSTR("SPB: Send date\r\n"));
		 #endif
	}
	else
	{
	  	sp_send_status(SP_ERR_IOERROR);
	}
}

/**
send the next file name from the current directory, for repository parsing
*/
void sp_send_nextfile(struct sp_states* p_state)
{
	SP_DEBUG_PUSH('|');
	FRESULT l_rc;
	FILINFO l_fileinfo;
	struct spvhd_nextfile* lp_nextfile;
	uint8_t i;
	
	l_fileinfo.fname[0] = ' ';
	l_fileinfo.fattrib = 0x0E;
	l_rc = FR_OK;
	// get the next file name but avoid hidden file, system file, volume label, and lfn entry
	while((l_fileinfo.fattrib & 0x0E) != 0 && l_rc == FR_OK && l_fileinfo.fname[0] != 0) 
		l_rc = f_readdir(&p_state->directory,&l_fileinfo);
	
/* File attribute bits for directory entry */ 
//#define	AM_RDO	0x01	/* Read only */ 
//#define	AM_HID	0x02	/* Hidden */ 
//#define	AM_SYS	0x04	/* System */ 
//#define	AM_VOL	0x08	/* Volume label */ 
//#define AM_LFN	0x0F	/* LFN entry */ 
//#define AM_DIR	0x10	/* Directory */ 
//#define AM_ARC	0x20	/* Archive */ 
	// requisite a buffer to send the file metadata
	//lp_nextfile = (struct spvhd_nextfile*)cset_new(0xFF,0xFFFFFFFF,0);
	// recycle the end of the buffer used for nibbled data within the packet object
	lp_nextfile = (struct spvhd_nextfile*)&g_sp_packet.data[SP_BUFDATSIZE - SPVHD_CFG_BLOCKSIZE]; 
	p_state->p_buffer = (uint8_t*)lp_nextfile;
	p_state->p_buffer_end = ((uint8_t*)(lp_nextfile))+sizeof(struct spvhd_nextfile);

	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_SND)	  printf_P(PSTR("SPB: filename start %02X\r\n"),l_fileinfo.fname[0]);
	#endif

	CSET_BANK_SET;
	if(l_rc == FR_OK && l_fileinfo.fname[0] != 0)
	{
		// size in KBytes
		lp_nextfile->filesize = (l_fileinfo.fsize>>10);
		// file attributes
		lp_nextfile->fileattr = l_fileinfo.fattrib;
		// and filename
		lp_nextfile->namesize = strlen(l_fileinfo.fname);
		for(i=0;i<lp_nextfile->namesize;i++)
			lp_nextfile->name[i]=l_fileinfo.fname[i];
		sprintf_P(&lp_nextfile->name[lp_nextfile->namesize],PSTR("%10lu"),lp_nextfile->filesize);
	}
	else
	{
	  	//sp_send_status(SP_ERR_IOERROR);
		lp_nextfile->filesize = 0;
		// file attributes
		lp_nextfile->fileattr = 0;
		// and filename
		lp_nextfile->namesize = 0;
		memset(&lp_nextfile->name[lp_nextfile->namesize],' ',10);
	}
	CSET_BANK_UNSET;
	
	sp_send_init(SP_PKTDAT,0,(uint8_t*)lp_nextfile,sizeof(struct spvhd_nextfile));
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_SND)	  printf_P(PSTR("SPB: Send nextfile %s\r\n"),lp_nextfile->name);
	#endif
}


/// Send a block parameter
/// @param p_state Pointer to the controle structure
/// @param l_type of parameter block requested
void sp_send_parm(struct sp_states* p_state, uint8_t l_statcode)
{
	SP_DEBUG_PUSH('|');
	struct spvhd_parms* lp_parms;
	uint8_t lp_pathname[22];
	uint8_t* lp_src;
	uint8_t* lp_dst;

	// determine the path to the configuration file
	strcpy_P(lp_pathname,PSTR("0:/spvhd/"));
	lp_dst = &lp_pathname[strlen(lp_pathname)];
	lp_src = PSTR("cfgboot.dat");
	strcpy_P(lp_dst,lp_src);
	
	// requisite a buffer to send the block
	//lp_parms = (struct spvhd_parms*)cset_new(0xFF,0xFFFFFFFF,0);
	// recycle the end of the buffer used for nibbled data within the packet object
	lp_parms = (struct spvhd_parms*)&g_sp_packet.data[SP_BUFDATSIZE - SPVHD_CFG_BLOCKSIZE]; 
	p_state->p_buffer = (uint8_t*)lp_parms;
	p_state->p_buffer_end = ((uint8_t*)(lp_parms))+sizeof(struct spvhd_parms);

	if( cfg_file_io(lp_pathname, p_state->devnum, 0, p_state->p_buffer) == 0)	
		sp_send_init(SP_PKTDAT,0,(uint8_t*)lp_parms,sizeof(struct spvhd_parms));
	else
	  	sp_send_status(SP_ERR_IOERROR);
	
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_SND)	  puts_P(PSTR("SPB: Send parms\r\n"));
	#endif
}
