/*
=============================================================================
Smartport Virtual Hard Drive, version 1.00.0

Copyright 2010-2012 Cédric PELTIER, pcedric@hotmail.com

This file is part of the Smartport VHD project

Smartport VHD is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 3

Smartport VHD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Smartport VHD. If not, see <http://www.gnu.org/licenses/>.
=============================================================================
*/

#include "spvhd.h"
#include "protocol.h"
#include "rwts.h"
#include <string.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#ifdef CFG_DEBUG_USB
#include "usb_serial.h"
#endif
#ifdef CFG_DEBUG_SERIAL
#if(LUFA_VERSION==100219)
#include <LUFA/Drivers/Peripheral/SerialStream.h>
#else
#include <LUFA/Drivers/Peripheral/Serial.h>
#endif
#endif


int16_t g_rwts_headpos = 0;
uint8_t g_rwts_index = 255;

/// Phases to sequence index translation
/// <- down to track 0
/// -> up to track 40
uint8_t g_rwts_idx[] = {255,0,2,1,4,255,3,255,6,7,255,255,5,255,255,255,255};
/// Previous sequence index by new sequence index matrix to alter head position
int8_t g_rwts_seq[] = {
	 0,	 1,	 2,	 0,	 0,	 0,	-2,	-1,
	-1,	 0,	 1,	 2,	 0,	 0,	 0,	-2,
	-2,	-1,	 0,	 1,	 2,	 0,	 0,	 0,
	 0,	-2,	-1,	 0,	 1,	 2,	 0,	 0,
	 0,	 0,	-2,	-1,	 0,	 1,	 2,	 0,
	 0,	 0,	 0,	-2,	-1,	 0,	 1,	 2,
	 2,	 0,	 0,	 0,	-2,	-1,	 0,	 1,
	 1,	 2,	 0,	 0,	 0,	-2,	-1,	 0
};

/// Translate lowest bits without any bit manipulation
uint8_t g_rwts_lookup62[] = {0x00,0x20,0x10,0x30, 0x00,0x08,0x04,0x0C, 0x00,0x02,0x01,0x03};
/// Translate 6 bits groups into 6 and 2 recognized nibbles
uint8_t g_rwts_nibble62[] = {0x96,0x97,0x9A,0x9B,0x9D,0x9E,0x9F,0xA6,0xA7,0xAB,0xAC,0xAD,0xAE,0xAF,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF,0xCB,0xCD,0xCE,0xCF,0xD3,0xD6,0xD7,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF,0xE5,0xE6,0xE7,0xE9,0xEA,0xEB,0xEC,0xED,0xEE,0xEF,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF};

void rwts_read_position()
{
	int16_t l_delta;
	uint8_t l_phases = PINF & 15;
	uint8_t l_index = g_rwts_idx[l_phases];
	if(l_index == 255)
		return;
	if(g_rwts_index == 255)
	{
		g_rwts_index = l_index;
		return;
	}
	l_delta = g_rwts_seq[((uint16_t)(g_rwts_index))*8 + ((uint16_t)(l_index))]; 
	if(l_delta == 0)
		return;
	g_rwts_index = l_index;
	g_rwts_headpos += (int16_t)l_delta;
	if(g_rwts_headpos<0)
		g_rwts_headpos = 0;
	else if(g_rwts_headpos>139)
		g_rwts_headpos = 139;
	
	#if defined(CFG_DEBUG_USB) || defined(CFG_DEBUG_SERIAL)
	if(gb_debug & CFG_DEBUG_MASK_RWTS)
	{	/*
		uint8_t c;
		if(l_phases<10)
			c = '0'+l_phases;
		else
			c = 'A'+l_phases-10;
		putchar(c);
		*/
		printf_P(PSTR(" %02X"),g_rwts_headpos);
	}
	#endif
}

/// Send a 256 bytes buffer as a GCR62 encoded sector to the host
/// @param l_volume Volume number to encode into the sector
/// @param l_track Track number to encode within the sector
/// @param l_sector Sector number to encode within the sector
/// @param lp_src Pointer to the source buffer
/// @param lp_dst Pointer to a buffer containing the encoded sector in return
///  at least 441 bytes must be allocated for this buffer
void rwts_send_sector(uint8_t l_volume,uint8_t l_track, uint8_t l_sector, uint8_t* lp_src, uint8_t* lp_dst)
{
	uint8_t i;
	uint8_t l_crc1 = 0;
	uint8_t l_crc2 = l_volume ^ l_track ^ l_sector;

	// Gap 1
	/*
	for(i=8;i!=0;i--)
	{
		memcpy_P(lp_dst,g_sp_sync,6);
		lp_dst+=6;
	}
	*/
	memcpy_P(lp_dst,g_sp_sync,6);
	lp_dst+=6;
	
	// We can start to send bit stream to the host
	// during the time we are encoding the sector
	
	*lp_dst = 0xD5;
	lp_dst++;
	*lp_dst = 0xAA;
	lp_dst++;
	*lp_dst = 0x96;
	lp_dst++;
	*lp_dst = GCR44E1(l_volume);
	lp_dst++;
	*lp_dst = GCR44E2(l_volume);
	lp_dst++;
	*lp_dst = GCR44E1(l_track);
	lp_dst++;
	*lp_dst = GCR44E2(l_track);
	lp_dst++;
	*lp_dst = GCR44E1(l_sector);
	lp_dst++;
	*lp_dst = GCR44E2(l_sector);
	lp_dst++;
	*lp_dst = GCR44E1(l_crc2);
	lp_dst++;
	*lp_dst = GCR44E2(l_crc2);
	lp_dst++;
	*lp_dst = 0xDE;
	lp_dst++;
	*lp_dst = 0xAA;
	lp_dst++;
	*lp_dst = 0xEB;
	lp_dst++;

	// Gap 2	
	memcpy_P(lp_dst,g_sp_sync,6);
	lp_dst+=6;
	
	*lp_dst = 0xD5;
	lp_dst++;
	*lp_dst = 0xAA;
	lp_dst++;
	*lp_dst = 0xAD;
	lp_dst++;
	
	// Encode 2 lowest bits from source into one destination byte, in descending order
	// require 256 / 3 => 86 bytes
	// on the avr platform simple loops with pointers are more efficients
	
	// Less efficient on the avr platform
	//for(i=85;i!=255;i--)
	//	lp_dst[i] = g_rwts_lookup62[ 8+(lp_src[172+i]&3) ] | g_rwts_lookup62[ 4+(lp_src[86+i]&3) ] | g_rwts_lookup62[ (lp_src[i]&3) ];
	// Encode this first part into recognized nibbles
	for(i=86;i!=0;i--,lp_dst++,lp_src++)
		*lp_dst = g_rwts_lookup62[ 8 + ( *lp_src&3 ) ];
	lp_dst -= 86;
	for(i=86;i!=0;i--,lp_dst++,lp_src++)
		*lp_dst |= g_rwts_lookup62[ 4 + ( *lp_src&3 ) ];
	lp_dst -= 86;
	for(i=84;i!=0;i--,lp_dst++,lp_src++)
		*lp_dst |= g_rwts_lookup62[ ( *lp_src&3 ) ];
	//l_crc2 = lp_dst[2];
	lp_dst -= 84;
	lp_src -= 256; // 86+86+84
	for(i=86;i!=0;i--,lp_dst++)
	{
		l_crc1 ^= *lp_dst;
		*lp_dst = g_rwts_nibble62[l_crc1];
	}
	
	// Encode 6 highest bits from source directly into recognized nibbles, in ascending order
	i=0;
	do
	{
		l_crc1 ^= (*lp_src >> 2);
		*lp_dst = g_rwts_nibble62[l_crc1];
		lp_dst++;
		lp_src++;
	}
	while(++i==0);

	*lp_dst = g_rwts_nibble62[l_crc1];
	lp_dst++;
	*lp_dst = 0xDE;
	lp_dst++;
	*lp_dst = 0xAA;
	lp_dst++;
	*lp_dst = 0xEB;
	lp_dst++;
	
	// Gap 3
	for(i=4;i!=0;i--)
	{
		memcpy_P(lp_dst,g_sp_sync,6);
		lp_dst+=6;
	}
}

// WRITE GAP ONE, WHICH CONTAINS 48 SELF-SYNC BYTES
// WRITE THE ADDRESS FIELD, WHICH CONTAINS:
//   - PROLOGUE (D5AA96)
//   - VOLUME NUMBER ("4 AND 4" ENCODED)
//   - TRACK NUMBER ("4 AND 4" ENCODED)
//   - SECTOR NUMBER ("4 AND 4" ENCODED)
//   - CHECKSUM ("4 AND 4" ENCODED)
//   - EPILOGUE (DEAAEB)
// WRITE GAP TWO, WHICH CONTAINS SIX SELF-SYNC BYTES
// WRITE THE DATA FIELD, WHICH CONTAINS:
//   - PROLOGUE (D5AAAD)
//   - 343 6-BIT BYTES OF NIBBLIZED DATA, INCLUDING A 6-BIT CHECKSUM
//   - EPILOGUE (DEAAEB)
// WRITE GAP THREE, WHICH CONTAINS 27 SELF-SYNC BYTES

/* DO logical order  0 1 2 3 4 5 6 7 8 9 A B C D E F */
/*    physical order 0 D B 9 7 5 3 1 E C A 8 6 4 2 F */

/* PO logical order  0 E D C B A 9 8 7 6 5 4 3 2 1 F */
/*    physical order 0 2 4 6 8 A C E 1 3 5 7 9 B D F */

